/*****************
 * "config.cpp"
 *
 *  Copyright 2014 Steve Shih
 *
 *  This file is part of Actuation.
 *
 *  Actuation is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Actuation is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Actuation.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Description:
 *
 *
 *****************/

/*
#include <iostream>
#include <stdio.h>
*/

#include "config.h"
#include "logging.h"

#define MAX_COMMAND_LENGTH 32
#define MAX_VALUE_LENGTH 32

using namespace std;
const char* Config::class_name_ = "Config";

Config::Config()
  :
  // Default Settings
  dispense_steps_(4),
  max_height_(550),
  max_width_(700),

  //max values
  max_steps_(30000),
  //SVG parsing settings
  //these work on all SVGs tested so far (that do not contain arcs)
  //TODO: store these in a config file or allow them to be changed within the program

  //all electrode coordinates are scaled by this value before performing intersection calculations
  // 5 works for most devices currently
  //decrease if not all intersections are found
  //increase if incorrect intersections are found
  multiple_(5),

  //these are similar are used in different cases
  //DO NOT CHANGE IF USING THE SAME CONNECTORS (the problem is elsewhere)
  tolerance_(2), //the max variation in the vertical position of the connectors
  precision_(1), //the max height difference among connectors

  prec_(0.1), //the amount by which to extend line segments when checking intersection
  div_(1), //used when checking whether the endpoints of line segments are the same

  debug_(0)
{
}

Config::~Config() {

}

bool Config::read(const char* profile_path) {
  const char* function_name = "read()";
  profile_path_ = profile_path;

  // default experiment dir is just the current directory
  experiment_logs_dir_ = boost::filesystem::current_path().string();

  std::ifstream config_file;
  std::string config_file_name = "actuation.conf";
  std::ostringstream msg;
  // if a config file exists in the current directory, it takes priority
  if(boost::filesystem::exists(config_file_name)) {
    msg << "using local config file="
        << (boost::filesystem::current_path()/config_file_name).string();
    config_file.open(config_file_name.c_str());
  } else if(boost::filesystem::exists(
                   (profile_path_/config_file_name).string().c_str())) {
    msg << "config file=" << (profile_path_/config_file_name).string();
    config_file.open((profile_path_/config_file_name).string().c_str());
  } else {
    msg << "no config file found";
  }
  Logging::LogDebug(msg.str().c_str(),class_name_,function_name);
  if(config_file.is_open()) {
    std::string line;
    boost::regex re_comment("^\\s*#");
    boost::regex re_parameter("^\\s*(\\S+)\\s*=\\s*(.+)\\s*$");
    boost::smatch matches;
    while(getline(config_file, line)) {
      std::ostringstream msg;
      msg << "Read config line: " << line;
      Logging::LogDebug(msg.str().c_str(),class_name_,function_name);
      //check if the line is a comment
      if(boost::regex_search(line,re_comment)) {
        Logging::LogDebug("This line is a comment",class_name_,function_name);
      } else if(boost::regex_search(line,matches,re_parameter)) {
        processCommand(matches[1],matches[2]);
      } else {
        Logging::LogError("Invalid syntax",class_name_,function_name);
      }
    }
    config_file.close();
    return true;
  }
  return false;
}

bool Config::processCommand(std::string command, std::string value) {
  const char* function_name = "processCommand";
  if(command.compare("dispense_steps") == 0) {
    dispense_steps_ = atoi(value.data());
    return true;
  } else if(command.compare("max_height") == 0) {
    max_height_ = atoi(value.data());
    return true;
  } else if(command.compare("max_width") == 0) {
    max_width_ = atoi(value.data());
    return true;
  } else if(command.compare("max_steps") == 0) {
    max_steps_ = atoi(value.data());
    return true;
  } else if(command.compare("multiple") == 0) {
    multiple_ = atof(value.data());
    return true;
  } else if(command.compare("tolerance") == 0) {
    tolerance_ = atof(value.data());
    return true;
  } else if(command.compare("precision") == 0) {
    precision_ = atof(value.data());
    return true;
  } else if(command.compare("prec") == 0) {
    prec_ = atof(value.data());
    return true;
  } else if(command.compare("div") == 0) {
    div_ = atof(value.data());
    return true;
  } else if(command.compare("experiment_logs_dir") == 0) {
    if(!boost::filesystem::exists(value.data())) {
      Logging::LogDebug("Experiment log directory does not exist. Create it",
                        class_name_,function_name);
      if(!boost::filesystem::create_directory(value.data())) {
        Logging::LogWarning("Create directory failed",
                          class_name_,function_name);
        return false;
      }
    }
    experiment_logs_dir_ = value.data();
    return true;
  } else if(command.compare("procedures_dir") == 0) {
    if(!boost::filesystem::exists(value.data())) {
      Logging::LogDebug("Procedures directory does not exist. Create it",
                        class_name_,function_name);
      if(!boost::filesystem::create_directory(value.data())) {
        Logging::LogWarning("Create directory failed",
                          class_name_,function_name);
        return false;
      }
    }
    procedures_dir_ = value.data();
    return true;
  } else if(command.compare("devices_dir") == 0) {
    if(!boost::filesystem::exists(value.data())) {
      Logging::LogDebug("Devices directory does not exist",
                        class_name_,function_name);
      return false;
    }
    devices_dir_ = value.data();
    return true;
  } else if(command.compare("device_path") == 0) {
    return set_device_path(value.data());
  } else if(command.compare("debug") == 0) {
    debug_ = atoi(value.data());
    return true;
  } else {
    Logging::LogWarning("Unsupported parameter",
                        class_name_,function_name);
  }
  return false;
}

bool Config::set_device_path(const char* device_path) {
  const char* function_name = "set_device_path";
  if(!boost::filesystem::exists(device_path)) {
    Logging::LogDebug("Device file does not exist",
                      class_name_,function_name);
    return false;
  }
  device_path_ = device_path;
  return true;
}
