#ifndef _DMF_CONTROLLER_H_
#define _DMF_CONTROLLER_H_

#ifndef AVR
  #include <iostream>
  #include <fstream>
  #include <string>
  #include <vector>
#endif
#include "SerialProtocol.h"


class DmfController : public SerialProtocol {
public:
  static const uint32_t BAUD_RATE = 115200;

  // AD5206 (digital potentiometer)
  //    TODO:
  //    Eventually, all of these variables should defined only on the arduino.
  //    The PC can interogate device using CMD_GET_NUMBER_OF_AD_CHANNELS,
  //    CMD_GET_POTENTIOMETER_MAX, CMD_GET_NUMBER_OF_POTENTIOMETER_STEPS
  static const uint16_t POTENTIOMETER_MAX = 100000;
  static const uint16_t NUMBER_OF_POTENTIOMETER_STEPS = 255;
  static const uint8_t NUMBER_OF_AD_CHANNELS = 2;
  static const uint8_t AD5206_SLAVE_SELECT_PIN = 10; // digital pot
  static const uint16_t MAX_SAMPLES = 100;
#ifdef AVR
  // I2C bus
  // =======
  // A4 SDA
  // A5 SCL

  // SPI bus
  // =======
  // D10 SS
  // D11 MOSI
  // D13 SCK

  // MAX7300 (gpio) chip addresses
  static const uint8_t MAX7300_FIRST_ROW = 0x40;
  static const uint8_t MAX7300_SECOND_ROW = 0x41;
  //need to place a third row here
  static const uint8_t MAX7300_THIRD_ROW = 0x44;
  static const uint8_t MAX7300_PORT_ADDRESSES[];

  // LTC6904 (programmable oscillator) chip address
  static const uint8_t LTC6904 = 0x17;

  static const float AMPLIFIER_GAIN = 100;
 // static const uint16_t NUMBER_OF_ELECTRODES = 40;
 static const uint16_t NUMBER_OF_ELECTRODES = 80;
#else
  static const char CSV_INDENT_[];
#endif

  //////////////////////////////////////////////////////////////////////////////
  //
  // Commands:
  //
  // Commands are uint8_t and should have the MSB=1 (replies will have MSB=0).
  //
  // Valid commands are in the range 0x80 to 0xFF.
  //
  // The following commands are reserved (they are defined by the
  // SerialProtocol class):
  //
  // 0x80 -> CMD_GET_PROTOCOL_NAME
  // 0x81 -> CMD_GET_PROTOCOL_VERSION
  // 0x82 -> CMD_ENABLE_CRC
  //
  //////////////////////////////////////////////////////////////////////////////

  // Accessors and mutators
  static const uint8_t CMD_GET_DEVICE_NAME =              0x83;
  static const uint8_t CMD_GET_DEVICE_VERSION =           0x84;
  static const uint8_t CMD_GET_NUMBER_OF_ELECTRODES =     0x85;
  static const uint8_t CMD_GET_NUMBER_OF_AD_CHANNELS =    0x86; //TODO
  static const uint8_t CMD_GET_POTENTIOMETER_MAX =        0x87; //TODO
  static const uint8_t CMD_GET_STATE_OF_ALL_ELECTRODES =  0x88;
  static const uint8_t CMD_SET_STATE_OF_ALL_ELECTRODES =  0x89;
  static const uint8_t CMD_GET_STATE_OF_ELECTRODE =       0x8A;
  static const uint8_t CMD_SET_STATE_OF_ELECTRODE =       0x8B;
  static const uint8_t CMD_GET_ACTUATION_VOLTAGE =        0x8C; //TODO
  static const uint8_t CMD_SET_ACTUATION_VOLTAGE =        0x8D; //TODO
  static const uint8_t CMD_GET_ACTUATION_FREQUENCY =      0x8E; //TODO
  static const uint8_t CMD_SET_ACTUATION_FREQUENCY =      0x8F; //TODO
  static const uint8_t CMD_GET_AVAILABLE_SAMPLING_RATES = 0x90;
  static const uint8_t CMD_SET_SAMPLING_RATE =            0x91;
  static const uint8_t CMD_SAMPLE_PHASE =                 0x92;
  // Other commands
  static const uint8_t CMD_SYSTEM_RESET =                 0xB0; //TODO
  static const uint8_t CMD_DEBUG_MESSAGE =                0xB1; //TODO
  static const uint8_t CMD_DEBUG_ON =                     0xB2; //TODO
  static const uint8_t CMD_SAMPLE_VOLTAGE =               0xB3;

  //////////////////////////////////////////////////////////////////////////////
  //
  // Return codes:
  //
  // Return codes are uint8_t.  Some generic return codes are defined by the
  // SerialProtocol class.  You may also create custom return codes as long as
  // they are > 0x10.
  //
  //////////////////////////////////////////////////////////////////////////////

  DmfController();
  ~DmfController();

// In our case, the PC is the only one sending commands
#ifndef AVR
  // Remote accessors (return code is from reply packet)
  uint8_t get_name(std::string& name);
  uint8_t get_version(std::string& version);
  uint8_t get_number_of_electrodes(uint16_t& number_of_electrodes);
  uint8_t get_state_of_all_electrodes(
                           std::vector<uint8_t>& state_of_electrodes);
  uint8_t get_state_of_electrode(const uint16_t electrode,
                                 uint8_t& state_of_electrode);
  uint8_t get_available_sampling_rates(std::vector<float>& sampling_rates);

  // Remote mutators (return code is from reply packet)
  uint8_t set_state_of_electrode(const uint16_t electrode, const uint8_t state);
//  uint8_t set_state_of_electrodes(std::vector<uint16_t> electrodes,
//                                  std::vector<uint8_t> states);
  uint8_t set_state_of_all_electrodes(const std::vector<uint8_t> state);
  uint8_t set_actuation_voltage(const float v_rms);
  uint8_t set_actuation_frequency(const float freq_hz);
  uint8_t set_sampling_rate(const uint8_t sampling_rate);

  // other functions
  uint8_t SampleVoltage(std::vector<uint16_t>& voltage_buffer,
                        std::vector<uint8_t> ad_channels,
                        std::vector<uint8_t> pot_steps,
                        uint8_t n_samples,
                        uint8_t n_sets=1,
                        uint16_t delay_between_sets_ms=0,
                        const std::vector<uint8_t> state =
                          std::vector<uint8_t>());
  uint8_t SamplePhase(std::vector<uint16_t>& phase_buffer);
  uint8_t SetExperimentLogFile(const char* file_name);
  void LogExperiment(const char* message);
  void LogExperimentV(const char* vel);
  float MillisecondsSinceLast();
#else
  void begin();
#endif

  // local accessors
  const char* protocol_name() { return PROTOCOL_NAME_; }
  const char* protocol_version() { return PROTOCOL_VERSION_; }
  const char* name() { return NAME_; } //device name
  const char* version() { return VERSION_; }// device version

private:
  // private static members
  static const char PROTOCOL_NAME_[];
  static const char PROTOCOL_VERSION_[];
  static const char NAME_[];
  static const char VERSION_[];

  // private functions
  virtual void ProcessCommand(uint8_t cmd);
  virtual void ProcessReply(uint8_t cmd);

#ifdef AVR
  void UpdateElectrode(const uint16_t electrode);
  void UpdateAllElectrodes();
  void SendI2C(uint8_t row, uint8_t cmd, uint8_t data);
  void SendSPI(uint8_t pin, uint8_t address, uint8_t data);
#else
  uint8_t SendCommand(const uint8_t cmd);
#endif

#ifdef AVR
  //private members
  uint8_t state_of_electrodes_[NUMBER_OF_ELECTRODES];
  uint16_t voltage_buffer_[MAX_SAMPLES];
#else
  std::string remote_name_;
  std::string remote_version_;
  std::vector<uint8_t> state_of_electrodes_;
  std::vector<uint16_t> voltage_buffer_;
  std::vector<uint16_t> phase_buffer_;
  std::vector<float> available_sampling_rates_;
  std::string experiment_log_file_name_;
  std::string vel_file_name_;
  std::ofstream experiment_log_file_;
  std::ofstream vel_file;
 // this also opens the file
  boost::posix_time::ptime t_last_;
#endif
};
#endif // _DMF_CONTROLLER_H_
