/*****************
 * "electrode.cpp"
 *
 *  Copyright 2014 Steve Shih
 *
 *  This file is part of Actuation.
 *
 *  Actuation is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Actuation is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Actuation.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Description:
 *
 *  Contains functions for creating, parsing, and modifying electrode classes
 *
 *****************/
#include "electrode.h"

using namespace std;

Electrode::Electrode()
  :
  type(-1),
  position(0),
  prev(NULL),
  next(NULL)
{
  shapes.push_back(new Line(-1, -1));
  state = new int[configuration.max_steps()];
  //initialize all states to GROUND
  for (int i = 0; i < configuration.max_steps(); i++)
    state[i] = GROUND;
}

//create an Electrode given an input string from an svg file
Electrode::Electrode(char * p)
  :
  type(0),
  prev(NULL),
  next(NULL)
{
  state = new int[configuration.max_steps()];
  int i;

  //fill in spaces with newlines and determine whether this is a closed Electrode
  for (i = 5; p[i]!='\"'; i++)
    {
      if (p[i]==' ')
	p[i]='\n';
      else if (p[i]=='z')
        {
	  type=1;
	  p[i]='\0';
	  p[i-1]='\0';
        }
    }
  p[i]='\0';
  p+=2;
  i=0;
  //determine the coordinates of the Electrode
  bool parsing_arc = false;
  bool parsing_line = false;
  bool parsing_curve = false;
  bool uppercase = false;
  bool radius_set = false;
  bool endpoint_set = false;
  double x, radx, endx, bx, ex;
  double y, rady, endy, by, ey;
  double flag1;
  int flag2, flag3;

  do
    {
      p++;
       if(*p == 'm'||*p == 'l')
	{
	  uppercase = false;
	  parsing_arc = false;
	  parsing_line = true;
	  parsing_curve = false;
	  p+=2;
	}
      else if(*p == 'M'||*p == 'L')
	{
	  parsing_arc = false;
	  parsing_line = true;
	  parsing_curve = false;
	  uppercase = true;
	  p+=2;
	}
      else if(*p == 'a')
	{
	  parsing_line = false;
	  parsing_arc = true;
	  parsing_curve = false;
	  uppercase = false;
	  radius_set = false;
	  endpoint_set = false;
	  flag1 = -1;
	  flag2 = -1, flag3 = -1;
	  p+=2;
	}
      else if(*p == 'A')
	{
	  parsing_line = false;
	  parsing_arc = true;
	  parsing_curve = false;
	  uppercase = true;
	  radius_set = false;
	  endpoint_set = false;
	  flag1 = -1;
	  flag2 = -1, flag3 = -1;
	  p+=2;
	}
      else if(*p == 'c')
	{
	  parsing_line = false;
	  parsing_arc = false;
	  parsing_curve = true;
	  uppercase = false;
	  p+=2;
	}
      if(parsing_line)
	{
	  sscanf(p, "%lf,%lf", &x, &y);
	  if (!uppercase&&size()>0)
	    {
	      shapes.push_back(new Line(x, y, shapes.back()));
	    }
	  else
	    {
	      shapes.push_back(new Line(x, y));
	    }
	}
      else if(parsing_curve)
	{
	  sscanf(p, "%lf,%lf\n%lf,%lf\n%lf,%lf", &bx, &by, &ex, &ey, &x, &y);
	  shapes.push_back(new Curve(bx, by, ex, ey, x, y, shapes.back()));
	  p=strchr(p,'\n'); //after beginning control point
	  p++;
	  p=strchr(p,'\n'); //after ending control point
	  p++;
	}
      else if(parsing_arc)
	{
	  //this seems to work
	  //there can be less that 3 flags but not in inkscape plain svg output
	  sscanf(p, "%lf,%lf\n%lf\n%d\n%d\n%lf,%lf", &radx, &rady, &flag1, &flag2, &flag3, &endx , &endy);

#ifndef _WIN32
      //doesn't compile on Windows?
	  shapes.push_back(new Arc(radx, rady, flag1, flag2, flag3, endx, endy, shapes.back()));
#endif

	  p=strchr(p,'\n'); //after radii
	  p++;
	  p=strchr(p,'\n'); //after first flag
	  p++;
	  p=strchr(p,'\n'); //after flag2
	  p++;
	  p=strchr(p,'\n'); //after flag3
	  p++;
	}
    }
  while (NULL!=(p=strchr(p,'\n')));

  //determine whether the start and end points match
  if(endpoints_match()){
    shapes.pop_back();
    type = 1;
  }

  //initialize all states to GROUND
  for (i = 0; i < configuration.max_steps(); i++)
    state[i] = GROUND;
}

//DESTRUCTOR!
Electrode::~Electrode()
{
  shapes.clear();
  if(state!=NULL)
    delete []state;
}


//add a new Electrode to a sorted linked list of electrodes
void Electrode::addElectrode (Electrode * n)
{
  //both electrodes are in the bottom row
  if (position == 0 && n -> position == 0)
    {
      //the new electrode goes after the current
      if (shapes[0]->getX() < n->shapes[0]->getX())
        {
	  //
	  if (next == NULL)
            {
	      next = n;
	      n->next = NULL;
	      n->prev = this;
	      return;
	      //end of the list
            }
	  else
            {
	      next->addElectrode(n);
            }
        }
      //the new electrode goes before
      else
        {
	  prev->next = n;
	  n->prev = prev;
	  n->next = this;
	  prev = n;
        }
    }
  //electrodes are in different rows
  else if (position == 0 && n -> position == 1)
    {
      if (next == NULL)
        {
	  next = n;
	  n->next = NULL;
	  n->prev = this;
	  return;
	  //end of the list
        }
      else
        {
	  next->addElectrode(n);
        }
    }
  else if (position == 1 && n -> position == 0)
    {
      prev->next = n;
      n->prev = prev;
      n->next = this;
      prev = n;
    }
  //both electrodes are in the top row
  else if (position == 1 && n -> position == 1)
    {
      if (shapes[0]->getX() > n->shapes[0]->getX())
        {
	  if (next == NULL)
            {
	      next = n;
	      n->next = NULL;
	      n->prev = this;
	      return;
	      //end of the list
            }
	  else
            {
	      next->addElectrode(n);
            }
        }
      else
        {
	  prev -> next = n;
	  n -> prev = prev;
	  n -> next = this;
	  prev = n;
        }
    }
}

// determine whether a point is in a polygon
int Electrode::check_connection(point p)
{
  int i, n, count = 0;
  //determine the number of vertices that the electrode has
  for (n = 0; n < size(); n++);

  //determine the number of sides that a horizontal ray extending from
  //the point intersects
  Path * current;
  Path * previous = shapes.back();

  for (i = 0; i < n; previous = shapes[i++])
    {
      current = shapes[i];
      double new_x, m, b;
      //a horizontal line will never intersect the horizontal ray
      if (equal(current->getY(), previous->getY()))
        {
	  new_x = -1;
        }
      else if (equal(current->getX(), previous->getX()))
        {
	  new_x = current->getX();
        }
      else
	{
	  m = (current->getY() - previous->getY())/(current->getX() - previous->getX());
	  b = current->getY() - m * current->getX();
	  new_x = (p.y - b)/m;
        }
      if (p.y > MIN(current->getY(), previous->getY())&&
	  p.y < MAX(current->getY(), previous->getY())&&
	  p.x < new_x)
        {
	  count++;
        }
    }
  return(count % 2);
}

//determine if two line segments intersect
int line_intersection(point p1, point q1, point p2, point q2)
{

  //slope
  double m1 = (p1.y - q1.y)/(p1.x - q1.x);
  double m2 = (p2.y - q2.y)/(p2.x - q2.x);
  //y-intercept
  double b1 = p1.y - m1 * p1.x;
  double b2 = p2.y - m2 * p2.x;
  point poi;

  //both lines are vertical
  if (equal(p1.x, q1.x) &&
      equal(p2.x, q2.x))
    {
      return 0;
    }
  else if (equal(p1.y, q1.y) &&
	   equal(p2.y, q2.y))
    {
      return 0;
    }
  else if (equal(p1.x, q1.x)) //the first line is vertical
    {
      poi.x = p1.x;
      poi.y = m2 * poi.x + b2;
    }
  else if (equal(p2.x, q2.x)) //the second line is vertical
    {
      poi.x = p2.x;
      poi.y = m1 * poi.x + b1;
    }
  else if (equal(p1.y, q1.y)) //the first line is horizontal
    {
      poi.y = p1.y;
      poi.x = (poi.y - b2)/m2;
    }
  else if (equal(p2.y, q2.y)) //the second line is horizontal
    {
      poi.y = p2.y;
      poi.x = (poi.y - b1)/m1;
    }
  else  //none of the lines are horizontal or  vertical
    {
      poi.x = (b1 - b2)/(m2 - m1);
      poi.y = m1 * (b1 - b2) / (m2 - m1) + b1;
    }
  if (poi.x <= MAX(p1.x, q1.x) &&
      poi.x >= MIN(p1.x, q1.x) &&
      poi.y <= MAX(p1.y, q1.y) &&
      poi.y >= MIN(p1.y, q1.y) &&
      poi.x <= MAX(p2.x, q2.x) &&
      poi.x >= MIN(p2.x, q2.x) &&
      poi.y <= MAX(p2.y, q2.y) &&
      poi.y >= MIN(p2.y, q2.y))
    return 1;
  return 0;
}
//determine if two line segments intersect (extend the line segments slightly)
int prec_line_intersection(point p1, point q1, point p2, point q2)
{

  double m1 = (p1.y - q1.y)/(p1.x - q1.x);
  double m2 = (p2.y - q2.y)/(p2.x - q2.x);
  double b1 = p1.y - m1 * p1.x;
  double b2 = p2.y - m2 * p2.x;
  point poi;

  //both lines are vertical
  if (equal(p1.x, q1.x) &&
      equal(p2.x, q2.x))
    {
      return 0;
    }
  else if (equal(p1.y, q1.y) &&
	   equal(p2.y, q2.y))
    {
      return 0;
    }
  else if (equal(p1.x, q1.x)) //the first line is vertical
    {
      poi.x = p1.x;
      poi.y = m2 * poi.x + b2;
    }
  else if (equal(p2.x, q2.x)) //the second line is vertical
    {
      poi.x = p2.x;
      poi.y = m1 * poi.x + b1;
    }
  else if (equal(p1.y, q1.y)) //the first line is horizontal
    {
      poi.y = p1.y;
      poi.x = (poi.y - b2)/m2;
    }
  else if (equal(p2.y, q2.y)) //the second line is horizontal
    {
      poi.y = p2.y;
      poi.x = (poi.y - b1)/m1;
    }
  else  //none of the lines are or horizontal vertical
    {
      poi.x = (b1 - b2)/(m2 - m1);
      poi.y = m1 * (b1 - b2) / (m2 - m1) + b1;
    }
  if (poi.x <= MAX(p1.x, q1.x) + configuration.prec() &&
      poi.x >= MIN(p1.x, q1.x) - configuration.prec() &&
      poi.y <= MAX(p1.y, q1.y) + configuration.prec() &&
      poi.y >= MIN(p1.y, q1.y) - configuration.prec() &&
      poi.x <= MAX(p2.x, q2.x) + configuration.prec() &&
      poi.x >= MIN(p2.x, q2.x) - configuration.prec() &&
      poi.y <= MAX(p2.y, q2.y) + configuration.prec() &&
      poi.y >= MIN(p2.y, q2.y) - configuration.prec())
    return 1;
  return 0;
}
//check if two wires intersect
int Electrode::check_wire_intersection(Electrode * line2)
{
  int p1i, p2i;
  for (p1i = 0; p1i < size(); p1i++)
    {
      for (p2i = 0; p2i < line2->size(); p2i++)
        {
	  //if the end points of any two line segments are touching
	  if (equal(shapes[p1i]->getX()/configuration.div(),
		    line2->shapes[p2i]->getX()/configuration.div())
	      &&
	      equal(shapes[p1i]->getY()/configuration.div(),
		    line2->shapes[p2i]->getY()/configuration.div()))
            {
	      return 1;
            }
	  //if any two lines intersect
	  if (p1i > 0 &&
	      p2i > 0 &&
	      prec_line_intersection(shapes[p1i-1] -> getPoint(),
				     shapes[p1i] -> getPoint(),
				     line2 -> shapes[p2i-1] -> getPoint(),
				     line2 -> shapes[p2i] -> getPoint()))
	    return 1;
        }
    }
  return 0;
}

//determine whether a line segment and a Electrode intersect
int Electrode::check_intersection(point p, point q)
{
  //if either endpoint of the line segment is inside the polygon then they intersect
  if (check_connection(p) || check_connection(q))
    return 1;

  int i, n = 0;
  //determine the number of vertices that the Electrode has
  for (n = 0; n < size(); n++);

  point p2, q2 = shapes.back()->getPoint();

  for (i = 0; i < n; q2 = shapes[i++]->getPoint())
    {
      p2 = shapes[i]->getPoint();
      if (line_intersection(p, q, p2, q2))
	return 1;
    }
  return 0;
}

//increment the state of an Electrode at a specified step
void Electrode::changeState(int current_step)
{
  if (state[current_step]!=0)
    state[current_step]=0;
  else
    state[current_step] += 1;
}


std::string Electrode::getNumberString()
{
  std::string numString = "";
  char buff[8];
  if(numberSet.empty())
    return numString;
  set<int>::iterator it = numberSet.begin();
  sprintf(buff, "%d", *(it));
  numString += buff;
  it++;
  for(; it != numberSet.end(); it++)
    {
      sprintf(buff, "%d", *(it));
      numString+=",";
      numString+= buff;
    }
  return numString;
}

//return the keyboard keys which will activate the Electrode
std::string Electrode::getCharString()
{
  std::string charString = "";
  if(numberSet.empty())
    return charString;
  set<int>::iterator it = numberSet.begin();
  int num = *it;
//  if(num < 21)
    if(num < 41 && num != 0)
     charString += (char)('A'+ num);
//    charString += (char)('a'+ num-1);
 // else if(*(numberSet.begin()) > 20)
    else if(*(numberSet.begin()) > 40)
   // charString += (char)('A'+num-21);
    charString += (char)('A'+num-41);
  it++;

  for(; it != numberSet.end(); it++)
    {
      charString+=",";
      num = *it;
 //     if(num < 21)
    if(num < 41 && num != 0)
	charString += (char)('A'+ num);
     // else if(num > 20)
	//charString += (char)('A'+num-21);
	  else if(num > 40)
	charString += (char)('A'+num-41);
    }

  return charString;
}


void Electrode::draw_selected(cairo_t * cr, double scale, point top_left)
{
  int i = 0;
  double x =  shapes[0]->getX() - top_left.x;
  double y =  shapes[0]->getY() - top_left.y;
  cairo_move_to(cr, scale*x, scale*y);

  for (i = 1; i < size(); i++)
    {
      shapes[i]->draw(cr, scale, top_left);
    }
  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_fill(cr);
}

void Electrode::draw(cairo_t * cr, double scale, point top_left, int cState)
{
  int i = 0;
  double x =  shapes[0]->getX() - top_left.x;
  double y =  shapes[0]->getY() - top_left.y;
  cairo_move_to(cr, scale*x, scale*y);

  for (i = 1; i < size(); i++)
    {
      shapes[i]->draw(cr, scale, top_left);
    }
  //determine the colour and fill the Electrode
  if (type>0)
    {
      cairo_close_path(cr);
      if (cState==0)
	cairo_set_source_rgb(cr, 0, 0, 1);
      else if (cState==1)
	cairo_set_source_rgb(cr, 1, (1.0*215)/255, 0);
      else
	cairo_set_source_rgb(cr, 1, 0, 0);
      cairo_fill(cr);
    }
}

void Electrode::print()
{
  cout<<"size: "<<size()<<"type "<<type<<" points ";
  for (int i = 0; i < size(); i++)
    cout<<"| x "<<shapes[i]->getX()<<" y "<<shapes[i]->getY();
  cout << endl;
}

point Electrode::find_top_left()
{
  point p, current;
  p.x = shapes[0]->getX();
  p.y = shapes[0]->getY();
  for(int i = 1; i < size(); i++)
    {
      if(shapes[i]->getType()=='p')
	{
	  current.x = shapes[i]->getX();
	  current.y = shapes[i]->getY();
	  if(current.x <= p.x)
	    p.x = current.x;
	  if(current.y <= p.y)
	    p.y = current.y;
	}
    }
  return p;
}

point Electrode::find_bottom_right()
{
  point p, current;
  p.x = shapes[0]->getX();
  p.y = shapes[0]->getY();
  for(int i = 1; i < size(); i++)
    {
      if(shapes[i]->getType()=='p')
	{
	  current.x = shapes[i]->getX();
	  current.y = shapes[i]->getY();
	  if(current.x >= p.x)
	    p.x = current.x;
	  if(current.y >= p.y)
	    p.y = current.y;
	}
    }
  return p;
}

point Electrode::find_bottom_left()
{
  point p, current;
  p.x = shapes[0]->getX();
  p.y = shapes[0]->getY();
  for(int i = 1; i < size(); i++)
    {
      if(shapes[i]->getType()=='p')
	{
	  current.x = shapes[i]->getX();
	  current.y = shapes[i]->getY();
	  if(current.x <= p.x &&
	     current.y >= p.y)
	    p = current;
	}
    }
  return p;
}

point Electrode::find_center()
{
  point c;
  int i, total = 1;
  c.x = shapes[0]->getX();
  c.y = shapes[0]->getY();
  for(i = 1; i < size(); i++){
    if(shapes[i]->getType()=='p')
      {
	c.x += shapes[i]->getX();
	c.y +=shapes[i]->getY();
	total ++;
      }
  }
  c.x /= total;
  c.y /= total;
  return c;
}

//Inkscape does not recognize some DXF paths as closed
//check if ANY 2 points match
//as some files had shapes with overlapping lines
bool Electrode::endpoints_match()
{
  for(int i = 0; i<size();i++)
    for(int j = i+4; j<size(); j++){
      if(((int)(shapes[i]->getX()+0.5))==((int)(shapes[j]->getX()+0.5)) &&
	 ((int)(shapes[i]->getY()+0.5))==((int)(shapes[j]->getY()+0.5)))
	return true;
    }
  return false;

}

//check whether two values are close together
//and can be considered equal
int equal(double a, double b)
{
  if(a + 0.2 > b &&
     a - 0.2 < b)
    return 1;
  return 0;
}
