/*****************
 *  "main.cpp"
 *
 *  Copyright 2014 Steve Shih
 *
 *  This file is part of Actuation.
 *
 *  Actuation is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Actuation is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Actuation.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************/

#define BOOST_FILESYSTEM_VERSION 3 // for boost 1.44, only
#define BOOST_FILESYSTEM_NO_DEPRECATED

#include "include.h"
#include "config.h"
#include "electrode_set.h"
#include "dmf_controller.h"
#include "windows.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "visa.h" //can download from National Instrument website
#include "visatype.h" //same as visa.h

#define REALTIME 1
#define PROCEDURE 0


//for function generator
#define	CHECK(func)	\
do {\
    ViStatus _s = (func);\
    if (_s < 0)\
    {\
	fprintf(stderr, "Error: %s returned %d\n", #func, _s);\
	exit(0);\
    }\
} while (0)

/* Specify the default address for function generator -- I use an Agilent 33210A */

#define DEFAULT_LOGICAL_ADDRESS "usb0::0x0957::0x1507::MY48012079::INSTR"

boost::filesystem::path profile_path;

// main data structure containing information about the electrodes
ElectrodeSet* electrodes;
Config configuration;

double thres_volt =0;  //global variable for feedback
int count_readings = 0; //to tell if this droplet moved
double scale;
double v_rms = 0;
double cal_array[40][40][40]; //to hold the threshold values
vector<uint16_t> phase_buffer; //to hold the phase values

float read_volt;
float read_volt_orig;

//variables for function generator
ViSession   viDefaultRM, Instrument;
ViRsrc      TxtAddress = DEFAULT_LOGICAL_ADDRESS;
ViUInt32	actual;



//to hold the temp string when it reads feedback file
string temp;
string temp_vrms;

// used to crop the display area to the edges of the microfluidics device
point top_left, bottom_right;

//the direction that the drawing has been moved
point drag;

int current_step;
int* duration;

// create file for calibration file
fstream feedback_file;
fstream copy_file;


// the electrode that has been selected (-1 if none)
int selected;
// this can be set to false to forbid the paint function from drawing
bool do_draw;
// keep track of whether a successful connection to the arduino has been made
bool connected;

// keep track of when a procedure is running
bool running;

// display number labels above the electrodes
bool do_show_numbers;
// display keyboard shortcuts
bool do_show_keys;

// objects used for communication with arduino
DmfController device;

// gtk initialization
GtkWidget* main_window;
GtkWidget* viewport;
GtkWidget* draw;

// right toolbar
GtkWidget* number_box;
GtkWidget* number_button;
GtkWidget* increment_button;
GtkWidget* run_button;
GtkWidget* dispense_duration_entry;
GtkWidget* dispense_button;
GtkWidget* h_split;
GtkWidget* v_split;
GtkWidget* device_info;
GtkWidget* experiment_id_label;

// feedback
GtkWidget* feedbackLabel;
GtkWidget* feedbackResistorCheckButon;
GtkWidget* feedbackResistorHBox;
GtkWidget* feedbackResistorUnits;
GtkWidget* feedbackResistorEntry;
GtkWidget* drivingVoltageResistorCheckButon;
GtkWidget* drivingVoltageResistorHBox;
GtkWidget* drivingVoltageResistorUnits;
GtkWidget* drivingVoltageResistorEntry;
GtkWidget* voltageDisplay;
GtkWidget* phaseDisplay;
GtkWidget* voltageCutoff;

GtkWidget* feedbackThreshold;
GtkWidget* feedbackThresholdHBox;
GtkWidget* feedbackThresholdEntry;

GtkWidget* CalibrationCheckButon;
GtkWidget* UsecalibrationCheckButon;

// actuation voltage/frequency
GtkWidget* voltageHBox;
GtkWidget* voltageLabel;
GtkWidget* voltageBox;
GtkWidget* voltageUnits;
GtkWidget* frequencyHBox;
GtkWidget* frequencyLabel;
GtkWidget* frequencyBox;
GtkWidget* frequencyUnits;
GtkWidget* set_voltage_and_frequency_button;

GtkWidget* menu_bar;

// repeat command
GtkWidget* repeat_window;
GtkWidget* repeat_button;
GtkWidget* repeat_from;
GtkWidget* repeat_to;
GtkWidget* repeat_iterations;
GtkWidget* repeat_from_entry;
GtkWidget* repeat_to_entry;
GtkWidget* repeat_iterations_entry;
GtkWidget* repeat_v_box;
GtkWidget* repeat_from_h_box;
GtkWidget* repeat_to_h_box;
GtkWidget* repeat_iterations_h_box;

// file menu and buttons
GtkWidget* file;
GtkWidget* file_menu;
GtkWidget* open_svg_button;
GtkWidget* open_dxf_button;
GtkWidget* save_procedure;
GtkWidget* open_procedure;
GtkWidget* quit_program;

// view menu
GtkWidget* view;
GtkWidget* view_menu;
GtkWidget* show_numbers_check;
GtkWidget* show_keys_check;

// mode menu
GtkWidget* mode;
GtkWidget* mode_menu;
GtkWidget* real_time_radio;
GtkWidget* procedure_radio;

// tools menu
GtkWidget* tools;
GtkWidget* tools_menu;
GtkWidget* repeat_tool;
GtkWidget* clear_tool;
GtkWidget* clear_all_tool;

// boxes for widget organization
GtkWidget* main_v_box;
GtkWidget* main_h_box;
GtkWidget* left_v_box;
GtkWidget* right_v_box;
GtkWidget* bottom_h_box;

// bottom toolbar
GtkWidget* step_label;
GtkWidget* prev_step_button;
GtkWidget* next_step_button;
GtkWidget* duration_box;
GtkWidget* ms_label;
GtkWidget* duration_button;
GtkWidget* durations_button;
GtkWidget* stop_button;
GtkWidget* resume_button;

// separators
GtkWidget* sep1;
GtkWidget* sep2;
GtkWidget* sep3;
GtkWidget* sep4;
GtkWidget* sep5;

// used for passing multiple pointers to a function
typedef struct {
  GtkWidget* data1;
  GtkWidget* data2;
  GtkWidget* data3;
  GtkWidget* data4;
  GtkWidget* data5;
  GtkWidget* data6;
  GtkWidget* data7;
} gpointers;

// forward declarations
int initialize(const char* filename);
void load_procedure(char* filename);
void write_procedure(char* filename);
void sample_feedback(vector<uint8_t> ad_channels,
                     vector<uint8_t> pot_steps,
                     uint8_t n_samples,
                     uint8_t n_sets=1,
                     uint16_t delay_between_sets_ms=0,
                     vector<uint8_t> state_of_electrodes=vector<uint8_t>());
int current_mode();
void feedback_tap(GtkWidget* widget, double vrms, vector <uint8_t> temp_electrodes);

void milliSleep(int mSeconds) {
#ifdef _WIN32
  Sleep(mSeconds);
#else
  usleep(mSeconds*1000);
#endif
}

/* set state (on the device) for all electrodes with this index */
void setElectrodeState(int index, int state) {
  for (int nIndex = 0; nIndex < electrodes->getNumberSetSizeI(index); nIndex++) {
    int num = electrodes->getNumber(index, nIndex);
    //if (num > 0 && num <= 40)
    if (num > 0 && num <= 80){
      device.set_state_of_electrode((num-1), state);
    }
  }
}

// updates every electrode
void update_device() {
  if (connected && (current_mode()==REALTIME || running)) {
    std::vector<uint8_t> state_of_electrodes;
 //   for (int n = 1; n <= 40 ; n++) {
    for (int n = 1; n <= 80 ; n++) {
      state_of_electrodes.push_back(electrodes->getStateN(n, current_step));  //get state of electrodes and place it into a vector
    }
    if(gtk_toggle_button_get_active((GtkToggleButton*)feedbackResistorCheckButon)
       ||gtk_toggle_button_get_active((GtkToggleButton*)drivingVoltageResistorCheckButon)) {
      vector<uint8_t> ad_channels;
      vector<uint8_t> pot_steps;
      if(gtk_toggle_button_get_active((GtkToggleButton*)feedbackResistorCheckButon)) {
        ad_channels.push_back(0);
        pot_steps.push_back((uint8_t)round(atof(gtk_entry_get_text((GtkEntry*)feedbackResistorEntry))
                            /DmfController::POTENTIOMETER_MAX
                            *DmfController::NUMBER_OF_POTENTIOMETER_STEPS));
      }
      if(gtk_toggle_button_get_active((GtkToggleButton*)drivingVoltageResistorCheckButon)) {
        ad_channels.push_back(1);
        pot_steps.push_back((uint8_t)round(atof(gtk_entry_get_text((GtkEntry*)drivingVoltageResistorEntry))
                            /DmfController::POTENTIOMETER_MAX
                            *DmfController::NUMBER_OF_POTENTIOMETER_STEPS));
      }
      sample_feedback(ad_channels,
                      pot_steps,
                      100,//ad_channels.size(), // n_samples
                      1,  // n_sets
                      10, // delay_between_sets_ms
                      state_of_electrodes);


    } else {

      device.set_state_of_all_electrodes(state_of_electrodes);
    }
  }
}

// handle mouse button clicks
static gboolean mouse_click_event(GtkWidget *widget,
                                  GdkEvent *event,
                                  gpointers * data) {
  if (event->type == GDK_BUTTON_PRESS) {
    Logging::LogMessageF(5,
			 "Mouse clicked: x%.0lf y%.0lf",
			 NULL,
			 "mouse_click_event",
			 ((GdkEventButton*) event)->x,
			 ((GdkEventButton*) event)->y);

    point p;
    int old_selected = selected;
    selected = -1;
    gtk_entry_set_text(GTK_ENTRY(data->data1), "");

    //determine the location of the drawing widget
    point draw_offset;
    draw_offset.x = (GTK_WIDGET(data->data2))->allocation.x;
    draw_offset.y = (GTK_WIDGET(data->data2))->allocation.y;
    //determine the position of the mouse click using the coordinate system used by the electrodes
    p.x = (1.0*((GdkEventButton*)event)->x - draw_offset.x)/scale + top_left.x - drag.x;
    p.y = (1.0*((GdkEventButton*)event)->y - draw_offset.y)/scale + top_left.y - drag.y;

    for (int i = 0; i < electrodes->size(); i++) {
      if (electrodes->getTypeI(i) > 0) {
        if (electrodes->checkConnectionI(i, p)) {
          // handle left button mouse clicks and ignore double clicks
          if (((GdkEventButton*)event)->button == 1) {
            electrodes->changeStateI(i, current_step);
            update_device();

          // select an electrode using the right mouse button
          } else if (((GdkEventButton*)event)->button == 3
                 &&(old_selected == -1||old_selected!=i)) {
            selected = i;
            if (selected > -1) {
              char buff[32];
              sprintf(buff, "%s", electrodes->getNumberString(selected).data());
              gtk_entry_set_text(GTK_ENTRY(data->data1), buff);
            }
          }
        }
      }
    }

    gtk_widget_queue_draw(widget);
	//gdk_window_process_all_updates();

    if(gtk_toggle_button_get_active((GtkToggleButton*)feedbackResistorCheckButon)) {
    std::vector<uint8_t> state_of_electrodes;
  //  for (int n = 1; n <= 40 ; n++) {
    for (int n = 1; n <= 80 ; n++) {
        state_of_electrodes.push_back(electrodes->getStateN(n, current_step));  }
    //get state of electrodes and place it into a vector
    feedback_tap(widget, v_rms, state_of_electrodes);

    read_volt = atof(gtk_entry_get_text((GtkEntry*)voltageBox));
    float read_freq = atof(gtk_entry_get_text((GtkEntry*)frequencyBox));
    //here we should check if the droplet moved
    while (read_volt <= 1.1 && (gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon))) { //this means that droplet didn't move with increased time, spo this will increase voltage

    v_rms = 0;
    //increase the voltage
    read_volt += 0.02;

    std::ostringstream string_init_volt;
    string_init_volt << "VOLTage " << read_volt << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
	CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));

    feedback_tap(widget, v_rms, state_of_electrodes);

    //change back to original voltage
   // std::ostringstream string_init_volt;
    string_init_volt << "VOLTage " << read_volt_orig << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
	CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));
}

    }

  }
  return FALSE;
}

void sample_feedback(vector<uint8_t> ad_channels,
                     vector<uint8_t> pot_steps,
                     uint8_t n_samples,
                     uint8_t n_sets,
                     uint16_t delay_between_sets_ms,
                     vector<uint8_t> state_of_electrodes) {
  vector<uint16_t> voltage_buffer;

  //to hold calibration reading for file
  std::ostringstream read_vrms;
  stringstream string_read;

  //close file if it's already open
  if (feedback_file.is_open()) {
  feedback_file.close();
  }

  if (copy_file.is_open()) {
  copy_file.close();
  }

  //open feedback file for I/O operation
  feedback_file.open ("feedback_file.csv", fstream::in | fstream::out | fstream::app);
  copy_file.open ("copy_file.csv", fstream::in | fstream::out | fstream::app);

  // collect feedback samples
  device.SampleVoltage(voltage_buffer,
                       ad_channels,
                       pot_steps,
                       n_samples,
                       n_sets,
                       delay_between_sets_ms,
                       state_of_electrodes);

  //vector<uint16_t> phase_buffer;

 // device.SamplePhase(phase_buffer);

  double phase = 0;
  double phase1 = 0;

  // need to convert dac
/*   for (int s=0; s<25; s++) {
   phase = ((float)phase_buffer[s]/1024) * 5; // this will convert it to voltage

   phase1 += (phase*180)/5; // this will convert it from voltage to phase

   }
   phase1 /= 25; //average phase from two samples

*/
  double dc_bias = 0;
  for(int i=0; i<n_samples; i++) {
    dc_bias += (float)voltage_buffer[(n_sets-1)*ad_channels.size()*n_samples
                                  +ad_channels.size()*i]/1024*5/n_samples;
  }
  v_rms = 0;
  for(int i=0; i<n_samples; i++) {
    v_rms += pow((float)voltage_buffer[(n_sets-1)*ad_channels.size()*n_samples
                                  +ad_channels.size()*i]/1024*5-dc_bias,2)
                                  /n_samples;
  }

  v_rms = sqrt(v_rms);
  std::ostringstream v_rms_str;
  v_rms_str << v_rms;
  gtk_label_set_text(GTK_LABEL(voltageDisplay),v_rms_str.str().c_str());

  std::ostringstream phase_str;
  phase_str << phase1 << " deg";
  gtk_label_set_text(GTK_LABEL(phaseDisplay),phase_str.str().c_str());

  //this is to get the calibration voltage data
  if(gtk_toggle_button_get_active((GtkToggleButton*)CalibrationCheckButon)){

         //which electrodes are "on"
         vector <uint8_t> temp_vector;

     //    for (int i=0; i<40; i++) { //this will check for which electrodes are on
            for (int i = 0; i <80; i++) {
             if (state_of_electrodes[i] == 1)
                {
                  temp_vector.push_back(i);
                }
         }

         //what did we sample
         cout << "ARDUINO reading of VRMS: " << v_rms << endl;


         if (temp_vector.size() == 1){ //if only one electrode is turned on
          cal_array[temp_vector[0]][0][0] = v_rms; // set value to the array

         //to check if electode voltage is in file
         int flag = 0;
         if (feedback_file.is_open()) {
            while (!feedback_file.eof()) {

                //int flag = 0;
                //this is to read the electrode
                getline(feedback_file, temp);
                cout << "Found: " << temp  << endl;
                getline(feedback_file,temp_vrms);  //this is to read the rms voltage

                //convert rms to string
                string_read << (int)temp_vector[0] << ",";
                cout << "Vrms from file: " << temp_vrms << endl;
                cout << "Electrode to calibrate: " << (int)temp_vector[0] << endl;
                cout << "Electrode (string read): " << string_read.str() << endl;

                //if we have the electrode - replace the vrms
                if (temp == string_read.str()) {
                 cout << "THIS IS GREAT!" << "   Electrode from file: " << temp << " and Electrode from vector: " << string_read.str() << endl;
                 //copy this to a new file
                 copy_file << temp << endl;
                 copy_file << v_rms << "," << endl;
                 flag = 1;
               //  break;
                }
                else if (temp == "50,50" && flag != 1) {
                 cout << "END OF FILE! " << endl;
                 copy_file << string_read.str() << endl;
                 copy_file << v_rms << "," << endl;
                 copy_file << "50,50" << endl;
                }

                else if (temp == "50,50" && flag == 1){
                   copy_file << temp << endl;
                }
                 // here we will write to replace the vrms
                else if (temp != "")
                {
                 cout << "NOT FOUND!!!" << endl;
                 copy_file << temp << endl;
                 copy_file << temp_vrms << endl;
                }

                //clear contents of string_read
                string_read.str("");
               } // end while loop we have reached eof.
            } // end feedback open loop

          copy_file.close();
          feedback_file.close();
          int delete_value = remove("feedback_file.csv");
          int rename_value = rename("copy_file.csv","feedback_file.csv");
          cout << "temp_vector_size: " << temp_vector.size() << endl;
          cout << "temp vector: " << (int)temp_vector[0] << endl;
          cout << "cal_array: " << cal_array[temp_vector[0]][0] << endl;

    } //end if statement for temp_vector.size() = 1
         else if (temp_vector.size() == 2) {

          cal_array[temp_vector[0]][temp_vector[1]][0] = v_rms;
          cal_array[temp_vector[1]][temp_vector[0]][0] = v_rms;

         //to check if electode voltage is in file
         int flag = 0;
         if (feedback_file.is_open()) {
            while (!feedback_file.eof()) {

                //int flag = 0;
                //this is to read the electrode
                getline(feedback_file, temp);
                cout << "Found: " << temp  << endl;
                getline(feedback_file,temp_vrms);  //this is to read the rms voltage

                //convert rms to string
                string_read << (int)temp_vector[0] << "," << (int)temp_vector[1];
                cout << "Vrms from file: " << temp_vrms << endl;
                cout << "Electrode to calibrate: " << (int)temp_vector[0] << " AND " << (int)temp_vector[1] << endl;
                cout << "Electrode (string read): " << string_read.str() << endl;

                //if we have the electrode - replace the vrms
                if (temp == string_read.str()) {
                 cout << "THIS IS GREAT!" << "   Electrode from file: " << temp << " and Electrode from vector: " << string_read.str() << endl;
                 //copy this to a new file
                 copy_file << temp << endl;
                 copy_file << v_rms << "," << endl;
                 flag = 1;
               //  break;
                }
                else if (temp == "50,50" && flag != 1) {
                 cout << "END OF FILE! " << endl;
                 copy_file << string_read.str() << endl;
                 copy_file << v_rms << "," << endl;
                 copy_file << "50,50" << endl;
                }

                else if (temp == "50,50" && flag == 1){
                   copy_file << temp << endl;
                }
                 // here we will write to replace the vrms
                else if (temp != "")
                {
                 cout << "NOT FOUND!!!" << endl;
                 copy_file << temp << endl;
                 copy_file << temp_vrms << endl;
                }

                //clear contents of string_read
                string_read.str("");
               } // end while loop we have reached eof.
            } // end feedback open loop

          copy_file.close();
          feedback_file.close();
          int delete_value = remove("feedback_file.csv");
          int rename_value = rename("copy_file.csv","feedback_file.csv");


          cout << "temp_vector_size: " << temp_vector.size() << endl;
          cout << "temp vector[0]: " << (int)temp_vector[0] << endl;
          cout << "temp vector[1]: " << (int)temp_vector[1] << endl;
          cout << "cal_array: " << cal_array[temp_vector[0]][temp_vector[1]] << endl;
          cout << "cal_array: " << cal_array[temp_vector[1]][temp_vector[0]] << endl;
         }

         else if (temp_vector.size() > 2) {

          cal_array[temp_vector[0]][temp_vector[1]][temp_vector[2]] = v_rms;
       //   cal_array[temp_vector[1]][temp_vector[0]] = v_rms;

         //to check if electode voltage is in file
         int flag = 0;
         if (feedback_file.is_open()) {
            while (!feedback_file.eof()) {

                //int flag = 0;
                //this is to read the electrode
                getline(feedback_file, temp);
                cout << "Found: " << temp  << endl;
                getline(feedback_file,temp_vrms);  //this is to read the rms voltage

                //convert rms to string
                string_read << (int)temp_vector[0] << "," << (int)temp_vector[1] << "," << (int)temp_vector[2];
                cout << "Vrms from file: " << temp_vrms << endl;
                cout << "Electrode to calibrate: " << (int)temp_vector[0] << " AND " << (int)temp_vector[1] << " AND " << (int)temp_vector[2] << endl;
                cout << "Electrode (string read): " << string_read.str() << endl;

                //if we have the electrode - replace the vrms
                if (temp == string_read.str()) {
                 cout << "THIS IS GREAT!" << "   Electrode from file: " << temp << " and Electrode from vector: " << string_read.str() << endl;
                 //copy this to a new file
                 copy_file << temp << endl;
                 copy_file << v_rms << "," << endl;
                 flag = 1;
               //  break;
                }
                else if (temp == "50,50" && flag != 1) {
                 cout << "END OF FILE! " << endl;
                 copy_file << string_read.str() << endl;
                 copy_file << v_rms << "," << endl;
                 copy_file << "50,50" << endl;
                }

                else if (temp == "50,50" && flag == 1){
                   copy_file << temp << endl;
                }
                 // here we will write to replace the vrms
                else if (temp != "")
                {
                 cout << "NOT FOUND!!!" << endl;
                 copy_file << temp << endl;
                 copy_file << temp_vrms << endl;
                }

                //clear contents of string_read
                string_read.str("");
               } // end while loop we have reached eof.
            } // end feedback open loop

          copy_file.close();
          feedback_file.close();
          int delete_value = remove("feedback_file.csv");
          int rename_value = rename("copy_file.csv","feedback_file.csv");


          cout << "temp_vector_size: " << temp_vector.size() << endl;
          cout << "temp vector[0]: " << (int)temp_vector[0] << endl;
          cout << "temp vector[1]: " << (int)temp_vector[1] << endl;
          cout << "temp vector[2]: " << (int)temp_vector[2] << endl;
          cout << "cal_array: " << cal_array[temp_vector[0]][temp_vector[1]][temp_vector[2]] << endl;
         }


      }

}

void feedback_tap(GtkWidget *widget, double vrms, vector <uint8_t> temp_electrodes)
{
    double threshold; //threshold voltage to decide if droplet is present.
    vector <uint8_t> temp_vector; //this is to hold which electrodes are on.
    vector<uint8_t> num_pot_steps; //same as pots_steps
    vector <uint16_t> voltage_array; //same as voltage buffer
    vector<uint8_t> ad_channels; //same as ad_channels
    uint8_t n_samples = 100; //set to 100 samples
    ad_channels.push_back(0); //set to only one channel

    int time_duration = atoi(gtk_entry_get_text((GtkEntry*)duration_box));
    int start_clock = (int)clock();
    int current_clock = (int)clock();
    double read_vrms;


      //close file if it's already open
    if (feedback_file.is_open()) {
    feedback_file.close();
    }

   //open feedback file for I/O operation
   feedback_file.open ("feedback_file.csv", fstream::in | fstream::out | fstream::app);


    int total_off_time; //the time to check how long the switch is off.

        for (int i=0; i<80; i++) { //this will check for which electrodes are on
             if (temp_electrodes[i] == 1)
                {
                    temp_vector.push_back(i); //save the electrodes that are on.
                }
        }

    //this is to distinguish beween droplet and dispensing
        if (temp_vector.size() == 2 && gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon))
        {
        //    if (cal_array[temp_vector[0]][temp_vector[1]] != NULL) {
                if (feedback_file.is_open()) {
                    while (!feedback_file.eof()) {


                        getline(feedback_file, temp);
                        getline(feedback_file,temp_vrms);

                        stringstream string_read;
                        string_read << (int)temp_vector[0] << "," << (int)temp_vector[1];

                        if (temp == string_read.str()) {
                           //change the string to double?
                           (istringstream)temp_vrms >> read_vrms;
                            threshold = read_vrms - 0.05;
                            cout << "Threshold from file: " << threshold << endl;
                }
            }
        }
    }

    else if (temp_vector.size() > 2 && gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon))
        {
                if (feedback_file.is_open()) {
                    while (!feedback_file.eof()) {


                        getline(feedback_file, temp);
                        getline(feedback_file,temp_vrms);

                        stringstream string_read;
                        string_read << (int)temp_vector[0] << "," << (int)temp_vector[1] << "," << (int)temp_vector[2];

                        if (temp == string_read.str()) {
                           //change the string to double?
                           (istringstream)temp_vrms >> read_vrms;
                            threshold = read_vrms - 0.05;
                            cout << "Threshold from file: " << threshold << endl;
                }
            }
        }
    }

        else if(temp_vector.size() == 1 && gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon)) {

              if (feedback_file.is_open()) {
                    while (!feedback_file.eof()) {
                        getline(feedback_file, temp);

                        getline(feedback_file,temp_vrms);
                        stringstream string_read;
                        string_read << (int)temp_vector[0] << ",";

                        if (temp == string_read.str()) {
                           //how to change the string to double?
                           (istringstream)temp_vrms >> read_vrms;
                            threshold = read_vrms - 0.05;
                            cout << "Threshold from file: " << threshold << endl;
        }
      }
    }
 }
        else {
        threshold = ((double)(atof(gtk_entry_get_text((GtkEntry*)feedbackThresholdEntry))));
        }
    //this will loop until droplet is covering electrode
    //reset count reading - we would only run this after 50 trials if we set a new voltage
    count_readings = 0;

    if ((read_volt > read_volt_orig) && gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon))
    {
     //threshold += (read_volt_orig * 0.45);  // this is the slope (calibration factor) when we increase voltage
    }

    while ((temp_vector.size() > 0) && (vrms < threshold) && count_readings < 50)
    {

    vrms = 0;  //reset to zero


     //now we will sample the voltage
     num_pot_steps.push_back((uint8_t)round(atof(gtk_entry_get_text((GtkEntry*)feedbackResistorEntry))
                            /DmfController::POTENTIOMETER_MAX
                            *DmfController::NUMBER_OF_POTENTIOMETER_STEPS));

     device.SampleVoltage(voltage_array,
                        ad_channels,
                       num_pot_steps,
                       n_samples,
                       1,
                       0,
                       temp_electrodes);


    for(int i=0; i<n_samples; i++) {
    vrms += pow((float)voltage_array[(1-1)*ad_channels.size()*n_samples
                                  +ad_channels.size()*i]/1024*5-2.5,2)
                                  /n_samples;
    }
    vrms = sqrt(vrms);
    std::ostringstream v_rms_str1;
    v_rms_str1 << vrms;
    gtk_label_set_text(GTK_LABEL(voltageDisplay),v_rms_str1.str().c_str());

    current_clock = (int)clock();
    count_readings = count_readings + 1;
    std::ostringstream msg;
    msg << "Feedback," << endl;
    msg << "total off time: " << total_off_time << " ms" << endl;
    msg << "time_feedback: " << current_clock - start_clock - total_off_time << " ms" << endl;
    msg << "time_duration: " << time_duration << " ms" << endl;
    msg << "time_left: " << time_duration - (current_clock - start_clock - total_off_time) << " ms" << endl;
    msg << "THRESHOLD: " << threshold << " V" << endl;
    cout << "THRESHOLD: " << threshold << " V" << endl;
    msg << "vrms " << vrms << " V" << endl;
    cout << "vrms: " << vrms << " V" << endl;
    cout << "counter: " << count_readings << endl;

    device.LogExperiment(msg.str().c_str());
    }
    thres_volt = threshold;
    v_rms = vrms;
}

// handle key presses
static gboolean key_press_event(GtkWidget* widget,
                                GdkEvent* event,
                                gpointer data) {
  if (event->type == GDK_KEY_PRESS) {
    switch(((GdkEventKey*)event)->keyval)
      {
      case GDK_Up:
	drag.y -= 20;
	break;
      case GDK_Down:
	drag.y += 20;
	break;
      case GDK_Left:
	drag.x -= 20;
	break;
      case GDK_Right:
	drag.x += 20;
	break;
      default:
	break;
      }


    Logging::LogMessageF(5,
			 "Key pressed: %c",
			 NULL,
			 "key_press_event",
			 ((GdkEventKey*)event)->keyval - GDK_a+'a');

    // lowercase
    if (gdk_keyval_is_lower(((GdkEventKey*)event)->keyval)) {
      int key = 1 + ((GdkEventKey*)event)->keyval - GDK_a;
   //   if (key > 0 && key < 21)
        if (key > 0 && key < 41)  {
        electrodes->changeStateN(key, current_step);
        if (connected&&current_mode() == REALTIME) {
          device.set_state_of_electrode(key-1, electrodes->getStateN(key, current_step));
        }
      }
    // uppercase
    } else {
     // int key = 21 + ((GdkEventKey*)event)->keyval - GDK_A;
      //if (key > 20 && key < 41)
       int key = 41 + ((GdkEventKey*)event)->keyval - GDK_A;
        if (key > 40 && key < 81) {
	electrodes->changeStateN(key, current_step);
        if (connected&&current_mode() == REALTIME) {
          device.set_state_of_electrode(key-1, electrodes->getStateN(key, current_step));
        }
      }
    }
  }
  gtk_widget_queue_draw(widget);
  return FALSE;
}

// zoom in and out
static gboolean scroll_callback(GtkWidget* widget,
                                GdkEventScroll* event,
                                gpointer data) {
  if (event->direction == GDK_SCROLL_UP) {
    scale*=1.04;
  } else {
    scale/=1.04;
  }
  gtk_widget_queue_draw(widget);

  Logging::LogMessage(8, " -- scrolling -- ", NULL, "scroll_callback");
  return FALSE;
}

// enable/disable different options in the VIEW menu
static gboolean show_numbers(GtkWidget* widget, gpointers* data) {
  if (do_show_numbers == true) {
    do_show_numbers = false;
  } else {
    do_show_numbers = true;
  }
  gtk_widget_queue_draw ((GtkWidget*)data);
  return FALSE;
}
static gboolean show_keys(GtkWidget* widget, gpointers* data) {
  if (do_show_keys == true) {
    do_show_keys = false;
  } else {
    do_show_keys = true;
  }
  gtk_widget_queue_draw ((GtkWidget*)data);
  return FALSE;
}

// REALTIME or PROCEDURE
int current_mode() {
  if(gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(real_time_radio))) {
    return REALTIME;
  }
  return PROCEDURE;
}

void set_current_mode(int mode) {
  if(mode==REALTIME) {
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(real_time_radio), TRUE);
  } else if(mode==PROCEDURE) {
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(procedure_radio), TRUE);
  }
  update_device();
}

void next_step_button_clicked(GtkWidget* widget, gpointers* data) {
  Logging::LogMessage(9, "", NULL, "next_step_button_clicked");

  // update the step inidicator
  if (current_step < configuration.max_steps() -1) {
    current_step++;
    update_device();

    char buff[16];
    sprintf(buff, "Step: #%d", current_step);
    gtk_label_set_text(GTK_LABEL(data ->data1), buff);
    sprintf(buff, "%d", duration[current_step]);
    gtk_entry_set_text(GTK_ENTRY(data -> data2), buff);
    gtk_widget_queue_draw (data -> data3);

  // display a warning if the last step is reached
  } else {
    GtkWidget* dialog = gtk_message_dialog_new (GTK_WINDOW(data->data3),
                            GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_WARNING,
                            GTK_BUTTONS_CLOSE,
                            "Already at last step!\nThe limit is: %d.",
                            configuration.max_steps()-1 );
    // close the dialog
    g_signal_connect_swapped (dialog, "response",
                              G_CALLBACK (gtk_widget_destroy),
                              dialog);
    gtk_dialog_run (GTK_DIALOG (dialog));
  }
}

void prev_step_button_clicked(GtkWidget* widget, gpointers* data) {
  Logging::LogMessage(8, "", NULL, "prev_step_button_clicked");

  if (current_step > 0) {
    // update the step indicator
    current_step--;
    update_device();
    char buff[16];
    sprintf(buff, "Step: #%d", current_step);
    gtk_label_set_text(GTK_LABEL(data -> data1), buff);
    sprintf(buff, "%d", duration[current_step]);
    gtk_entry_set_text(GTK_ENTRY(data -> data2), buff);
    gtk_widget_queue_draw (data -> data3);
  } else {

    // display a warning if it is already step 1
    GtkWidget* dialog = gtk_message_dialog_new (GTK_WINDOW(data->data3),
                                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                                GTK_MESSAGE_WARNING,
                                                GTK_BUTTONS_CLOSE,
                                                "Already at first step!");
    // close the dialog
    g_signal_connect_swapped (dialog, "response",
                              G_CALLBACK (gtk_widget_destroy),
                              dialog);
    gtk_dialog_run (GTK_DIALOG (dialog));
  }
}

void open_svg_button_clicked(GtkWidget* widget, gpointer data) {
  Logging::LogDebug("", NULL, "open_svg_button_clicked");
  GtkWidget* dialog;
  dialog = gtk_file_chooser_dialog_new ("Open SVG File",
                                        (GtkWindow*)data,
                                        GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);
  boost::filesystem::path p = configuration.devices_dir();
  p /= "*";
  gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog),p.string().c_str());
  if (gtk_dialog_run (GTK_DIALOG (dialog))==GTK_RESPONSE_ACCEPT) {
    char* cfilename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (dialog));
    string filename = cfilename;
    free(cfilename);
    initialize(filename.c_str());
  }
  gtk_widget_destroy (dialog);
  gtk_widget_queue_draw ((GtkWidget*)data);
}

void open_dxf_button_clicked(GtkWidget* widget, gpointer data) {
  Logging::LogDebug("", NULL, "open_dxf_button_clicked");
  GtkWidget* dialog;
  dialog = gtk_file_chooser_dialog_new ("Open DXF File",
                                        (GtkWindow*)data,
                                        GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
    char* cdxf_filename =
                      gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    string dxf_filename = cdxf_filename;
    free(cdxf_filename);
    string svg_filename;

    std::ostringstream msg;
    msg << "Filename: " << dxf_filename;
    Logging::LogDebug(msg.str().c_str(),NULL,"open_dxf_button_clicked");

    size_t ext = dxf_filename.find(".dxf");
    if(ext==-1) {
      Logging::LogError("Not a DXF file!", NULL, "open_dxf_button_clicked");
      return;
    } else {
      svg_filename = dxf_filename.substr(0,ext-1);
      svg_filename += ".svg";
    }

    std::ostringstream export_arg;
    // run inkscape to convert the file
#ifdef _WIN32
    // set the command line arguments
    export_arg << " \"" << dxf_filename << "\" --export-plain-svg=\""
               << svg_filename << ".svg" << "\"";

    // setup a new process
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    memset(&pi, 0, sizeof(pi)); // Set up memory block
    memset(&si, 0 , sizeof(si)); // Set up memory block
    si.cb = sizeof(si); // Set structure size

    // if using a 64 bit system the location of inkscape will be different, try both
    if (CreateProcess("C:\\Program Files\\Inkscape\\inkscape.exe",
                      (char*)export_arg.str().c_str(),
                      NULL,NULL,NULL,NULL,NULL,NULL,&si,&pi)==0) {
        CreateProcess("C:\\Program Files (x86)\\Inkscape\\inkscape.exe",
                      (char*)export_arg.str().c_str(),
                      NULL,NULL,NULL,NULL,NULL,NULL,&si,&pi);
    }
    WaitForSingleObject(pi.hProcess, INFINITE);
#else // if using UNIX
    int status;
    // set the command line arguments
    export_arg << "--export-plain-svg=\"" << svg_filename;
    if (fork()==0) {
      execlp("inkscape\0", "inkscape\0",
             dxf_filename.c_str(), export_arg.str().c_str(), NULL);
      cout<<"SHOULD NEVER GET HERE"<<endl;
      exit(1);
    } else {
      wait(&status);
    }
#endif
  }
  gtk_widget_destroy (dialog);
  gtk_widget_queue_draw ((GtkWidget*)data);
}

void open_procedure_clicked(GtkWidget* widget, gpointer data) {
  GtkWidget* dialog;
  dialog = gtk_file_chooser_dialog_new ("Open Procedure File",
                                        (GtkWindow*)data,
                                        GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);
  boost::filesystem::path p = configuration.procedures_dir();
  p /= "*";
  gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog),p.string().c_str());
  if (gtk_dialog_run (GTK_DIALOG (dialog))==GTK_RESPONSE_ACCEPT) {
    set_current_mode(PROCEDURE);
    char* filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    load_procedure(filename);
    free (filename);
  }
  gtk_widget_destroy (dialog);
  gtk_widget_queue_draw (widget);
}

void save_procedure_clicked(GtkWidget* widget, gpointer data) {
  GtkWidget* dialog;
  dialog = gtk_file_chooser_dialog_new("Write File",
                                       (GtkWindow*)data,
                                       GTK_FILE_CHOOSER_ACTION_SAVE,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                       GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                       NULL);
  boost::filesystem::path p = configuration.procedures_dir();
  p /= "*";
  gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog),p.string().c_str());
  if (gtk_dialog_run (GTK_DIALOG (dialog))==GTK_RESPONSE_ACCEPT) {
    char* filename;
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    write_procedure(filename);
    free(filename);
  }
  gtk_widget_destroy (dialog);
  gtk_widget_queue_draw (widget);
}

void dispense_button_clicked(GtkWidget* widget, gpointers* data) {
  if (selected<0) {
    GtkWidget* dialog = gtk_message_dialog_new (GTK_WINDOW(data->data2),
                                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                                GTK_MESSAGE_WARNING,
                                                GTK_BUTTONS_CLOSE,
                                                "No electrode selected.\n"
                                                "Use the right mouse button to "
                                                "select an electrode.");
    // close the dialog
    g_signal_connect_swapped(dialog, "response",
                              G_CALLBACK (gtk_widget_destroy),
                              dialog);
    gtk_dialog_run(GTK_DIALOG (dialog));
  } else {
    // find the duration of each step during dispense

    int dispense_duration = atoi(gtk_entry_get_text((GtkEntry*)(data->data1)));
    Logging::LogMessageF(5,
			"Dispense duration: %d ms",
			NULL,
			"dispense_button_clicked",
			dispense_duration);

    int* dispense_electrodes = new int[configuration.dispense_steps()];
    dispense_electrodes[0] = selected;
    selected = -1;

    // find the numbers of the electrodes to dispense
    for (int i = 1;
	 i < configuration.dispense_steps() && dispense_electrodes[i-1] != -1; i++){
	   dispense_electrodes[i] = electrodes->getClosestI(i-1,
							    dispense_electrodes);
    }

    // run the dispense procedure
    for (int i = 0; i < configuration.dispense_steps() && dispense_electrodes[i] != -1; i++) {
      electrodes->setStateI(HIGH, dispense_electrodes[i], current_step);
      if (i!=0) {
        electrodes->setStateI(GROUND, dispense_electrodes[i-1], current_step);
      }
      if (connected) {
        setElectrodeState(dispense_electrodes[i], HIGH);
        if (i!=0) {
          setElectrodeState(dispense_electrodes[i-1], GROUND);
        }
      }
      if (i==configuration.dispense_steps()-1) {
        electrodes->setStateI(HIGH, dispense_electrodes[0], current_step);
        if (connected) {
          setElectrodeState(dispense_electrodes[0], HIGH);
        }
      }
      gtk_widget_queue_draw(data->data2);
      gdk_window_process_all_updates();
      milliSleep(dispense_duration);
    }
  }
}

void split(gpointers* data, int direction) {
  if (selected < 0) {
    GtkWidget* dialog = gtk_message_dialog_new (GTK_WINDOW(data->data2),
                                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                                GTK_MESSAGE_WARNING,
                                                GTK_BUTTONS_CLOSE,
                                                "No electrode selected.\n"
                                                "Use the right mouse button "
                                                "to select an electrode.");
    // close the dialog
    g_signal_connect_swapped (dialog, "response",
                              G_CALLBACK(gtk_widget_destroy),
                              dialog);
    gtk_dialog_run (GTK_DIALOG(dialog));
  } else {
    // find the duration of each step during dispense
    int dur = atoi(gtk_entry_get_text((GtkEntry*)(data->data1)));
    electrodes->setStateI(HIGH, selected, current_step);
    if (connected) {
      setElectrodeState(selected, HIGH);
    }
    int _selected = selected;
    selected = -1;
    gtk_widget_queue_draw(data->data2);
    gdk_window_process_all_updates();
    milliSleep(dur);

    int first;
    int second;
    if (direction == 0) {
      first = electrodes->get_left(_selected);
      second = electrodes->get_right(_selected);
    } else {
      first = electrodes->get_top(_selected);
      second = electrodes->get_bottom(_selected);
    }
    electrodes->setStateI(GROUND, _selected, current_step);
    if (connected) {
      setElectrodeState(_selected, GROUND);
    }
    if (first != -1) {
      electrodes->setStateI(HIGH, first, current_step);
      if (connected) {
        setElectrodeState(first, HIGH);
      }
    }
    if (second!=-1) {
      electrodes->setStateI(HIGH, second, current_step);
      if (connected) {
        setElectrodeState(second, HIGH);
      }
    }
    gtk_widget_queue_draw(data->data2);
    gdk_window_process_all_updates();
  }
}

void h_split_clicked(GtkWidget* window, gpointers* data) {
  split(data, 0);
}

void v_split_clicked(GtkWidget* window, gpointers* data) {
  split(data, 1);
}

// run a procedure on the device
gboolean run_next_step(gpointer data) {
  char buff[16];
 GtkWidget* widget;

  if (duration[current_step]>-1
    &&current_step<configuration.max_steps()
    &&running) {

    Logging::LogMessageF(5,
			 "Writing step: %d",
			 NULL,
			 "run_next_step",
			 current_step);

    if (current_step>=0) {
      update_device();

      // we need to check for feedback
        if(gtk_toggle_button_get_active((GtkToggleButton*)feedbackResistorCheckButon)) {
            std::vector<uint8_t> state_of_electrodes;
                //changed here from 40 to 80
                for (int n = 1; n <= 80 ; n++) {
                    state_of_electrodes.push_back(electrodes->getStateN(n, current_step));  }
                    //get state of electrodes and place it into a vector
                    feedback_tap(widget, v_rms, state_of_electrodes);

                    read_volt = atof(gtk_entry_get_text((GtkEntry*)voltageBox));
                    float volt_dummy = read_volt;
                    float read_freq = atof(gtk_entry_get_text((GtkEntry*)frequencyBox));
                    //here we should check if the droplet moved
                    while ((v_rms < thres_volt) && read_volt <= 1.1 && (gtk_toggle_button_get_active((GtkToggleButton*)UsecalibrationCheckButon))) { //this means that droplet didn't move with increased time, spo this will increase voltage

                    v_rms = 0;
                    //increase the voltage
                    read_volt += 0.02;

                    std::ostringstream string_init_volt;
                    string_init_volt << "VOLTage " << read_volt << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
                    CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));

                    feedback_tap(widget, v_rms, state_of_electrodes);

                    //std::ostringstream string_init_volt;
                    string_init_volt << "VOLTage " << volt_dummy << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
                    CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));
                        }


                    }
    }

    sprintf(buff, "Step: #%d", current_step);
    gtk_label_set_text(GTK_LABEL(((gpointers*)data)->data1), buff);
    sprintf(buff, "%d", duration[current_step]);
    gtk_entry_set_text(GTK_ENTRY(((gpointers*)data)->data2), buff);
    gtk_widget_queue_draw(((gpointers*)data)->data3);
    gdk_window_process_all_updates();

    // if the stop button has been pressed, the next step will not run
    if (running) {
      // this function is only exact if the program is idle during the timeout time

      current_step++;
      g_timeout_add(duration[current_step], run_next_step, data);
    }
  } else {
    current_step--;
    running = false;
    sprintf(buff, "Step: #%d", current_step);
    gtk_label_set_text(GTK_LABEL(((gpointers*)data)->data1), buff);
    sprintf(buff, "%d", duration[current_step]);
    gtk_entry_set_text(GTK_ENTRY(((gpointers*)data)->data2), buff);
    gtk_widget_queue_draw(((gpointers*)data)->data3);
    gdk_window_process_all_updates();
  }
  return FALSE;
}

void run_button_clicked(GtkWidget* widget, gpointers* data) {
  // start a new file
  boost::filesystem::path exp_dir = configuration.experiment_logs_dir();
  for(int i=0; true; i++) {
    std::ostringstream experiment_id;
    experiment_id << "experiment " << i;
    gtk_label_set_text(GTK_LABEL(experiment_id_label),
                       experiment_id.str().c_str());
    experiment_id << ".csv";
    // find the first new file (or one that has a zero size
    if(boost::filesystem::exists(exp_dir/experiment_id.str())==false
       || boost::filesystem::file_size(exp_dir/experiment_id.str())==0) {
      device.SetExperimentLogFile((exp_dir/experiment_id.str()).string().c_str());
      break;
    }
  }
  running = true;
  current_step = 0;
  char buff[16];
  sprintf(buff, "Step: #%d", current_step);
  gtk_label_set_text(GTK_LABEL(data -> data1), buff);
  sprintf(buff, "%d", duration[current_step]);
  gtk_entry_set_text(GTK_ENTRY(data -> data2), buff);
  gtk_widget_queue_draw(data->data3);
  gdk_window_process_all_updates();
  run_next_step(data);
}

// stop a running procedure
void stop_button_clicked(GtkWidget* widget, gpointer* data) {
  if (!running) {
    return;
  }
  running = false;
}

void resume_button_clicked(GtkWidget* widget, gpointers* data) {
  running = true;
  // prevent from updating twice
  if (current_step!=0) {
    update_device();
  }
  run_next_step(data);
}

void set_voltage_and_frequency_button_clicked(GtkWidget* widget, gpointers* data) {

    float initial_volt = atof(gtk_entry_get_text((GtkEntry*)voltageBox));
    float initial_freq = atof(gtk_entry_get_text((GtkEntry*)frequencyBox));
 // device.set_actuation_frequency(freq_hz*1000);


    if (initial_volt < 1.5) {

    std::ostringstream string_init_volt;
    string_init_volt << "VOLTage " << initial_volt << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
	CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));

    std::ostringstream string_init_freq;
    string_init_freq << "FREQ " << initial_freq << "\n"; /* Set the frequency in Hz) */
	CHECK(viWrite(Instrument, (ViByte*)string_init_freq.str().c_str(), (ViUInt32)string_init_freq.str().size(), &actual));
    }
    else {

    initial_volt = 1.0;

    std::ostringstream string_init_volt;
    string_init_volt << "VOLTage " << initial_volt << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
	CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));

    std::ostringstream string_init_freq;
    string_init_freq << "FREQ " << initial_freq << "\n"; /* Set the frequency in Hz) */
	CHECK(viWrite(Instrument, (ViByte*)string_init_freq.str().c_str(), (ViUInt32)string_init_freq.str().size(), &actual));
    }
 /*
  float v_rms = atof(gtk_entry_get_text((GtkEntry*)voltageBox));
  device.set_actuation_voltage(v_rms);
  float freq_hz = atof(gtk_entry_get_text((GtkEntry*)frequencyBox));
  device.set_actuation_frequency(freq_hz*1000);*/
}

void duration_button_clicked(GtkWidget* widget, gpointer data) {
  duration[current_step] = atoi(gtk_entry_get_text((GtkEntry*)data));
}

void durations_button_clicked(GtkWidget* widget, gpointer data) {
  int new_dur = atoi(gtk_entry_get_text((GtkEntry*)data));
  for (int i = current_step; i < configuration.max_steps(); i++) {
    duration[i] = new_dur;
  }
}

void number_button_clicked(GtkWidget* widget, gpointers* data) {
  if (selected>-1) {
    string new_number = gtk_entry_get_text(GTK_ENTRY(data->data1));
    electrodes->changeNumberI(selected, new_number);
    gtk_widget_queue_draw (data->data3);
  }
}

void increment_button_clicked(GtkWidget* widget, gpointers* data) {
  // only increment if a connector is selected
  if (selected>-1&&electrodes->getTypeI(selected)==CONNECTOR) {
    int increment = atoi(gtk_entry_get_text(GTK_ENTRY(data->data1)));

    // determine which row to increment
    int row = electrodes->getPositionI(selected);

    // increment that row
    electrodes->incrementRow(row, increment);
    gtk_widget_queue_draw(data->data3);
  }
}

void show_repeat_window(GtkWidget* widget, GtkWidget* repeat_window) {
  gtk_widget_show_all(repeat_window);
}

void hide_repeat_window(GtkWidget* widget, GtkWidget* repeat_window) {
  gtk_widget_hide_all(repeat_window);
}

void repeat_button_clicked(GtkWidget* widget, gpointers* data) {
  int from_step = atoi(gtk_entry_get_text(GTK_ENTRY(data->data1)));
  int to_step = atoi(gtk_entry_get_text(GTK_ENTRY(data->data2)));
  int number_of_iterations = atoi(gtk_entry_get_text(GTK_ENTRY(data->data3)));
  for (int i=current_step;
       i<current_step+(1+to_step-from_step)*number_of_iterations;
       i++) {
    electrodes->copy_step(from_step + (i - current_step)%(1 + to_step - from_step) , i);
    duration[i] = duration[from_step + (i - current_step)%(1 + to_step - from_step)];
  }
  char buff[16];
  sprintf(buff, "%d", duration[current_step]);
  gtk_entry_set_text(GTK_ENTRY(data -> data5), buff);
  gtk_widget_queue_draw (data->data4);
}

void clear_current_step(GtkWidget* widget, gpointers* data) {
  electrodes->clear_step(current_step);
  duration[current_step] = 0;
  update_device();
  char buff[16];
  sprintf(buff, "%d", duration[current_step]);
  gtk_entry_set_text(GTK_ENTRY(data -> data1), buff);
  gtk_widget_queue_draw (data->data2);
}

void clear_all_steps(GtkWidget* widget, gpointers* data) {
  electrodes->clear_all();
  for (int i = 0; i< configuration.max_steps(); i++) {
    duration[i] = 0;
  }
  update_device();
  char buff[16];
  sprintf(buff, "%d", duration[current_step]);
  gtk_entry_set_text(GTK_ENTRY(data -> data1), buff);
  gtk_widget_queue_draw (data->data2);
}

// paint fuction
static gboolean on_expose_event(GtkWidget* widget,
                                GdkEventExpose* event,
                                gpointer data) {
  if (do_draw) {
    cairo_t* cr;
    cr = gdk_cairo_create(widget->window);

    // paint it black
    cairo_rectangle(cr,
                    0,
                    0,
                    configuration.max_width(),
                    configuration.max_height());
    cairo_stroke_preserve(cr);
    cairo_set_source_rgb(cr, 0, 0, 0);
    cairo_fill(cr);

    // settings
    cairo_set_line_width(cr, 0.4);
    // cairo_select_font_face(cr, "Courier New",
    //                      CAIRO_FONT_SLANT_NORMAL,
    //                       CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, 10);

    // drawing
    point temp;
    temp.x = top_left.x - drag.x;
    temp.y = top_left.y - drag.y;
    electrodes->draw(cr,
                     scale,
                     temp,
                     current_step,
                     selected,
                     do_show_numbers,
                     do_show_keys);
    cairo_destroy(cr);
  }
  //gtk_window_resize ((GtkWindow*)data,1, 1);
  return FALSE;
}

int initialize(const char* filename) {
  if(!boost::filesystem::exists(filename)) {
    std::ostringstream msg;
    msg << "File does not exist: " << filename;
    Logging::LogError(msg.str().c_str(),NULL,"initialize");
    return -1;
  }
  // parse the svg file
  ifstream fp(filename);
  if (!(fp.is_open())) {
    std::ostringstream msg;
    msg << "Failed to open: " << filename;
    Logging::LogError(msg.str().c_str(),NULL,"initialize");
    return -1;
  } else {
    // initialization
    do_draw = false;
    drag.x = 0;
    drag.y = 0;

    // count the number of lines in the file
    int size = 0;
    string line;
    while(getline(fp, line)) {
      size++;
    }
    fp.close();
    ifstream fp(filename);
    if (!(fp.is_open())){
      std::ostringstream msg;
      msg << "Failed to open: " << filename;
      Logging::LogError(msg.str().c_str(),NULL,"initialize");
      return -1;
    }

    // delete electrodes;
    Electrode** array = new Electrode*[size];
    delete electrodes;
    int i = 0;
    current_step = 0;
    selected = -1;
    int j;
    for (j=0; j<configuration.max_steps(); j++) {
      duration[j]=0;
    }
    char* p = NULL;
    char* q = NULL;
    // ignore everything before the g tag
    while(getline(fp, line)&&(p=strstr((char*)line.data(), "<g"))==NULL) {
      // skip these lines
    }
    p = NULL;

    while (getline (fp, line)) {
      // only read lines which define paths
      if ((p = strstr((char*)line.data(), "d=\"m"))!=NULL) {
        array[i++]=new Electrode(p);
        p = NULL;
      }

      // if the "M" is uppercase use absolute coordinates
      if ((q = strstr((char*)line.data(), "d=\"M"))!=NULL) {
          array[i++]=new Electrode(q);
          q = NULL;
      }
    }
    fp.close();

    array[i]=new Electrode();// blank electrode
    electrodes = new ElectrodeSet(array);

    // find the dimensions of the chip
    top_left = electrodes->find_top_left();
    bottom_right = electrodes->find_bottom_right();
    scale = min(configuration.max_height()/(bottom_right.y-top_left.y), configuration.max_width()/(bottom_right.x-top_left.x));
    do_draw = true;
  }
  update_device();
  configuration.set_device_path(filename);
  string title = "Actuation - ";
  title += filename;
  gtk_window_set_title(GTK_WINDOW(main_window),title.c_str());
  return 0;
}

// opens a previously saved procedure
void load_procedure(char* filename) {
  ifstream fp (filename);
  if (!(fp.is_open())) {
    std::ostringstream msg;
    msg << "Failed to open: " << filename;
    Logging::LogError(msg.str().c_str(),NULL,"load_procedure");
  } else {
    // load the procedure
    string cLine;
    // skip first two lines as they will indicate rows and columns
    getline(fp, cLine);
    getline(fp, cLine);
    int line = 0, step = 0;
    while ( getline (fp, cLine)) {
      if (line%2 == 0) {
        sscanf(cLine.data(), "%d", &(duration[step]));
      } else {
	      int current_electrode = 1;
	      for (int current_char = 0;
             cLine.data()[current_char]!='\n';
             current_char ++) {
          switch(cLine.data()[current_char]) {
            case '2':
              electrodes->setStateI(HIGH,
				    electrodes->getIndexN(current_electrode),
				    step);
              current_electrode++;
              break;
            case '1':
              electrodes->setStateI(GROUND,
				    electrodes->getIndexN(current_electrode),
				    step);
              current_electrode++;
              break;
            case '0':
              current_electrode++;
              break;
          }
        }
        step++;
      }
      line++;
    }
    duration[step]=-1;
    fp.close();
  }
}

// write the steps to a tab delimited text file to be read by LabVIEW
void write_procedure (char* filename) {
  ofstream output (filename);
  if (!(output.is_open())) {
    std::ostringstream msg;
    msg << "Failed to open: " << filename;
    Logging::LogError(NULL,"write_procedure");
  }

  unsigned short i = 0, j;
  // find the number of electrodes
  int num_electrodes = electrodes->last_num();

  // first two lines are the dimensions of the microfluidics chip
  output<<num_electrodes<<endl<<"1"<<endl;

  // the end of the instruction occurs when a step has negative duration
  for (i = 0; duration[i]>-1&&i<configuration.max_steps(); i++) {
    output<<duration[i]<<endl;

    // write the states for the current step
    for (j = 1; j <= num_electrodes; j++) {
      // print the tab delimited array representing the state
      output<<(electrodes->getStateN(j, i)+1)<<"\t";
    }
    output<<endl;
  }
  output.close();
}

// function called when quitting the program
gint clean_up (gpointer data) {
  Logging::LogMessage(5, "Cleaning up", NULL, "clean_up");
  delete electrodes;
  delete [] duration;
  return 0;
}

int main(int argc, char* argv[]) {



  const char* function_name = "main()";
#ifdef _WIN32
  char buffer[MAX_PATH];
  SHGetSpecialFolderPath(NULL, buffer, CSIDL_PERSONAL, 0);
  profile_path /= buffer;
#else
  profile_path = getenv("HOME");
#endif
  profile_path /= ".actuation";

  std::ostringstream msg;
  msg << "profile directory=" << profile_path.string();
  Logging::LogDebug(msg.str().c_str(),NULL,function_name);

  if(!boost::filesystem::exists(profile_path)) {
    Logging::LogDebug("The profile directory not exist. Create it",
                      NULL,function_name);
    if(boost::filesystem::create_directory(profile_path)==false) {
      Logging::LogWarning("Could not create profile directory",
                        NULL,function_name);
    }
  }

  // load the config file or use the default configuration
  configuration.read(profile_path.string().c_str());

  // if we're not in debug mode, set the log to only warnings and errors
  if (!configuration.debug()) Logging::SetLogLevel(4);

  duration = new int[configuration.max_steps()];

  connected = false;
  running = false;

  drag.x = 0;
  drag.y = 0;

  electrodes = new ElectrodeSet(NULL);
  gtk_init(&argc, &argv);

  // main_window settings
  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(main_window), "Actuation");
  gtk_window_set_default_size(GTK_WINDOW(main_window), 0, 0);

  // repeat tool window settings
  repeat_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(repeat_window), "Repeat Tool");
  gtk_window_set_default_size(GTK_WINDOW(repeat_window), 50, 50);

  // initialization of elements
  main_v_box = gtk_vbox_new(FALSE, 0);
  main_h_box = gtk_hbox_new(FALSE, 0);
  left_v_box = gtk_vbox_new(FALSE, 0);
  right_v_box = gtk_vbox_new(FALSE, 0);
  voltageHBox = gtk_hbox_new(FALSE, 0);
  frequencyHBox = gtk_hbox_new(FALSE, 0);
  bottom_h_box = gtk_hbox_new(FALSE, 10);
  repeat_v_box = gtk_vbox_new(FALSE, 0);
  repeat_from_h_box = gtk_hbox_new(FALSE, 0);
  repeat_to_h_box = gtk_hbox_new(FALSE, 0);
  repeat_iterations_h_box = gtk_hbox_new(FALSE, 0);
  draw = gtk_drawing_area_new();
  viewport = gtk_viewport_new(NULL, NULL);

  // intialize repeat tool elements
  repeat_from = gtk_label_new("From Step #:");
  repeat_from_entry = gtk_entry_new();
  repeat_to = gtk_label_new("To Step #:");
  repeat_to_entry = gtk_entry_new();
  repeat_iterations = gtk_label_new("Iterations:");
  repeat_iterations_entry = gtk_entry_new();
  repeat_button = gtk_button_new_with_label("Repeat");

  // initialize the menus
  menu_bar = gtk_menu_bar_new();
  file_menu = gtk_menu_new();
  view_menu = gtk_menu_new();
  mode_menu = gtk_menu_new();
  tools_menu = gtk_menu_new();
  file = gtk_menu_item_new_with_label("File");
  view = gtk_menu_item_new_with_label("View");
  mode = gtk_menu_item_new_with_label("Mode");
  tools = gtk_menu_item_new_with_label("Tools");

  open_procedure = gtk_menu_item_new_with_label("Open Procedure");
  save_procedure = gtk_menu_item_new_with_label("Save Procedure");
  quit_program = gtk_menu_item_new_with_label("Quit");
  open_svg_button = gtk_menu_item_new_with_label("Open SVG");
  open_dxf_button = gtk_menu_item_new_with_label("Open DXF");
  repeat_tool = gtk_menu_item_new_with_label("Repeat Steps");
  clear_tool = gtk_menu_item_new_with_label("Clear Step");
  clear_all_tool = gtk_menu_item_new_with_label("Clear All Steps");

  show_numbers_check = gtk_check_menu_item_new_with_label("Show Numbers");
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_numbers_check), TRUE);
  show_keys_check = gtk_check_menu_item_new_with_label("Show Keyboard Shortcuts");
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_keys_check), TRUE);
  do_show_numbers = true;
  do_show_keys = true;

  real_time_radio = gtk_radio_menu_item_new_with_label(NULL, "Real Time Mode");
  GSList* group = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(real_time_radio));
  procedure_radio = gtk_radio_menu_item_new_with_label(group, "Procedure Mode");
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(real_time_radio), TRUE);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), open_svg_button);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), open_dxf_button);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), open_procedure);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), save_procedure);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), quit_program);
  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), show_numbers_check);
  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), show_keys_check);
  gtk_menu_shell_append(GTK_MENU_SHELL(mode_menu), real_time_radio);
  gtk_menu_shell_append(GTK_MENU_SHELL(mode_menu), procedure_radio);
  gtk_menu_shell_append(GTK_MENU_SHELL(tools_menu), repeat_tool);
  gtk_menu_shell_append(GTK_MENU_SHELL(tools_menu), clear_tool);
  gtk_menu_shell_append(GTK_MENU_SHELL(tools_menu), clear_all_tool);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file), file_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(view), view_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(mode), mode_menu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tools), tools_menu);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), tools);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), view);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), mode);
  gtk_box_pack_start(GTK_BOX(main_v_box), menu_bar, FALSE, FALSE, 0);

  // initialize the toolbar buttons

  /*** Procedure Control ***/
  step_label = gtk_label_new("Step: #0");
  prev_step_button = gtk_button_new_with_label("Previous");
  next_step_button = gtk_button_new_with_label("Next");
  duration_box = gtk_entry_new();
  duration_button = gtk_button_new_with_label("Set Duration");
  durations_button = gtk_button_new_with_label("Set All");
  ms_label = gtk_label_new("(ms)");
  run_button = gtk_button_new_with_label("Actuate");
  stop_button = gtk_button_new_from_stock(GTK_STOCK_STOP);
  resume_button = gtk_button_new_with_label("Resume");

  /*** Arduino ***/
  device_info = gtk_label_new("Not connected");
  experiment_id_label = gtk_label_new("");

  /*** Actuation voltage/frequency ***/
  voltageBox = gtk_entry_new();
  gtk_entry_set_text((GtkEntry*)voltageBox,"0.1");
  voltageLabel = gtk_label_new("Voltage (Vpp)");
  voltageUnits  = gtk_label_new("V");
  frequencyBox = gtk_entry_new();
  gtk_entry_set_text((GtkEntry*)frequencyBox,"8000");
  frequencyLabel = gtk_label_new("Frequency");
  frequencyUnits = gtk_label_new("Hz");
  set_voltage_and_frequency_button = gtk_button_new_with_label("Set");

  /*** Feedback ***/
  feedbackLabel = gtk_label_new("Feedback");
  feedbackResistorCheckButon = gtk_check_button_new_with_label("Feedback Resistor");
  feedbackResistorHBox = gtk_hbox_new(FALSE, 0);
  feedbackResistorEntry = gtk_entry_new();
  gtk_entry_set_text((GtkEntry*)feedbackResistorEntry,"5000");
  feedbackResistorUnits = gtk_label_new("Ω");
  drivingVoltageResistorCheckButon = gtk_check_button_new_with_label("Driving Voltage");
  drivingVoltageResistorHBox = gtk_hbox_new(FALSE, 0);
  drivingVoltageResistorEntry = gtk_entry_new();
  gtk_entry_set_text((GtkEntry*)drivingVoltageResistorEntry,"1000");
  drivingVoltageResistorUnits = gtk_label_new("Ω");
  voltageDisplay = gtk_label_new("---");
  phaseDisplay = gtk_label_new("--- deg");
  voltageCutoff = gtk_entry_new();

  //threshold labels
  feedbackThreshold = gtk_label_new("Threshold voltage");
  feedbackThresholdHBox = gtk_hbox_new(FALSE, 0);
  feedbackThresholdEntry = gtk_entry_new();
  gtk_entry_set_text((GtkEntry*)feedbackThresholdEntry,"0.0");

  CalibrationCheckButon = gtk_check_button_new_with_label("Calibration?");
  UsecalibrationCheckButon = gtk_check_button_new_with_label("Use Calibration?");
  /*** Electrode Number Editing ***/
  number_box = gtk_entry_new();
  number_button = gtk_button_new_with_label("Change Number");
  increment_button = gtk_button_new_with_label("Increment Row");

  /*** Dispense and Split ***/
  dispense_duration_entry = gtk_entry_new();
  dispense_button = gtk_button_new_with_label("Dispense");
  h_split = gtk_button_new_with_label("Horizontal Split");
  v_split = gtk_button_new_with_label("Vertical Split");

  /*** Widget Separators ***/
  sep1 = gtk_hseparator_new();
  sep2 = gtk_hseparator_new();
  sep3 = gtk_hseparator_new();
  sep4 = gtk_hseparator_new();
  sep5 = gtk_hseparator_new();

  // drawing area settings
  gtk_widget_set_app_paintable(draw, TRUE);

  /*** Construct The Window ***

   main_v_box:
   ___________________________________________________________________
   menu_bar
   ___________________________________________________________________
   main_h_box:   | left_v_box:                          | right_v_box:
                 | draw                                 | buttons
		 | bottom_h_box: buttons and labels     | and
		 |                                      | labels
		 _____________________________________________________


   */

  gtk_widget_set_size_request(draw,
			      configuration.max_width(),
			      configuration.max_height());
  gtk_widget_set_size_request(prev_step_button, 90, 32);
  gtk_widget_set_size_request(step_label, 120, 32);
  gtk_widget_set_size_request(next_step_button, 90, 32);
  gtk_widget_set_size_request(duration_box, 50, 32);
  gtk_widget_set_size_request(run_button, 90, 32);
  gtk_widget_set_size_request(voltageBox, 90, -1);
  gtk_widget_set_size_request(frequencyBox, 90, -1);

  gtk_container_add(GTK_CONTAINER(main_window), main_v_box);
  gtk_container_add(GTK_CONTAINER(viewport), draw);
  //hScrollbar = gtk_hscrollbar_new(gtk_viewport_get_hadjustment(GTK_VIEWPORT(viewport)));
  //vScrollbar = gtk_vscrollbar_new(gtk_viewport_get_vadjustment(GTK_VIEWPORT(viewport)));
  //make a table with scrollbars

  gtk_box_pack_start(GTK_BOX(main_v_box), main_h_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_h_box), left_v_box, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(main_h_box), right_v_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(left_v_box), viewport, TRUE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(left_v_box), bottom_h_box, TRUE, FALSE, 0);

  /*** Arduino ***/
  gtk_box_pack_start(GTK_BOX(right_v_box), device_info, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), experiment_id_label, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), sep1, TRUE, FALSE, 15);

  /*** Actuation voltage/frequency ***/
  gtk_box_pack_start(GTK_BOX(voltageHBox), voltageLabel, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(voltageHBox), voltageBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(voltageHBox), voltageUnits, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(frequencyHBox), frequencyLabel, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(frequencyHBox), frequencyBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(frequencyHBox), frequencyUnits, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), voltageHBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), frequencyHBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), set_voltage_and_frequency_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), sep2, TRUE, FALSE, 15);

  /*** Feedback ***/
  gtk_box_pack_start(GTK_BOX(right_v_box), feedbackLabel, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), feedbackResistorCheckButon, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(feedbackResistorHBox), feedbackResistorEntry, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(feedbackResistorHBox), feedbackResistorUnits, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), feedbackResistorHBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), drivingVoltageResistorCheckButon, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(drivingVoltageResistorHBox), drivingVoltageResistorEntry, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(drivingVoltageResistorHBox), drivingVoltageResistorUnits, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), drivingVoltageResistorHBox, TRUE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(right_v_box), feedbackThreshold, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), feedbackThresholdHBox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(feedbackThresholdHBox), feedbackThresholdEntry, TRUE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(right_v_box), CalibrationCheckButon, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), UsecalibrationCheckButon, TRUE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(right_v_box), voltageDisplay, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), phaseDisplay, TRUE, FALSE, 0);
//  gtk_box_pack_start(GTK_BOX(right_v_box), voltageCutoff, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), sep3, TRUE, FALSE, 15);

  /*** Electrode Number Editing ***/
  gtk_box_pack_start(GTK_BOX(right_v_box), number_box, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), number_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), increment_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), sep4, TRUE, FALSE, 15);

  /*** Dispense and Split ***/
  gtk_box_pack_start(GTK_BOX(right_v_box), dispense_duration_entry, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), dispense_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), h_split, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), v_split, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(right_v_box), sep5, TRUE, FALSE, 15);

  /*** Procedure Control ***/
  gtk_box_pack_start(GTK_BOX(right_v_box), run_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), prev_step_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), step_label, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), next_step_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), duration_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), ms_label, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), duration_button, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(bottom_h_box), durations_button, TRUE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(bottom_h_box), stop_button, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(bottom_h_box), resume_button, TRUE, FALSE, 0);

  // construct the repeat tool window
  gtk_container_add(GTK_CONTAINER(repeat_window), repeat_v_box);
  gtk_box_pack_start(GTK_BOX(repeat_v_box), repeat_from_h_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_v_box), repeat_to_h_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_v_box), repeat_iterations_h_box, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_v_box), repeat_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_from_h_box), repeat_from, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_from_h_box), repeat_from_entry, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_to_h_box), repeat_to, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_to_h_box), repeat_to_entry, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_iterations_h_box), repeat_iterations, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(repeat_iterations_h_box), repeat_iterations_entry, TRUE, FALSE, 0);

  // events
  gtk_widget_add_events(GTK_WIDGET(main_window), GDK_BUTTON_PRESS_MASK);

  // structs for passing several data pointers at the same time
  gpointers data, data2, data4, data5, data6;
  data.data1 = step_label;
  data.data2 = duration_box;
  data.data3 = main_window;

  data2.data1 = number_box;
  data2.data2 = draw;
  data2.data3 = main_window;

  data4.data1 = dispense_duration_entry;
  data4.data2 = main_window;

  data5.data1 = repeat_from_entry;
  data5.data2 = repeat_to_entry;
  data5.data3 = repeat_iterations_entry;
  data5.data4 = main_window;
  data5.data5 = duration_box;

  data6.data1 = duration_box;
  data6.data2 = main_window;

  // menu buttons
  g_signal_connect(G_OBJECT(quit_program), "activate",
		   G_CALLBACK(gtk_main_quit), NULL);
  g_signal_connect(G_OBJECT(open_svg_button), "activate",
		   G_CALLBACK(open_svg_button_clicked), main_window);
  g_signal_connect(G_OBJECT(open_dxf_button), "activate",
		   G_CALLBACK(open_dxf_button_clicked), main_window);
  g_signal_connect(G_OBJECT(open_procedure), "activate",
		   G_CALLBACK(open_procedure_clicked), NULL);
  g_signal_connect(G_OBJECT(save_procedure), "activate",
		   G_CALLBACK(save_procedure_clicked), NULL);
  g_signal_connect(G_OBJECT(show_numbers_check), "activate",
		   G_CALLBACK(show_numbers), main_window);
  g_signal_connect(G_OBJECT(show_keys_check), "activate",
		   G_CALLBACK(show_keys), main_window);
  g_signal_connect(G_OBJECT(repeat_tool), "activate",
		   G_CALLBACK(show_repeat_window), repeat_window);
  g_signal_connect(G_OBJECT(repeat_button), "clicked",
		   G_CALLBACK(repeat_button_clicked), &data5);
  g_signal_connect(G_OBJECT(clear_tool), "activate",
		   G_CALLBACK(clear_current_step), &data6);
  g_signal_connect(G_OBJECT(clear_all_tool), "activate",
		   G_CALLBACK(clear_all_steps), &data6);

  // button signal handlers
  g_signal_connect(G_OBJECT(prev_step_button), "clicked",
		   G_CALLBACK(prev_step_button_clicked), &data);
  g_signal_connect(G_OBJECT(next_step_button), "clicked",
		   G_CALLBACK(next_step_button_clicked), &data);
  g_signal_connect(G_OBJECT(duration_button), "clicked",
		   G_CALLBACK(duration_button_clicked), (gpointer) duration_box);
  g_signal_connect(G_OBJECT(durations_button), "clicked",
		   G_CALLBACK(durations_button_clicked), (gpointer) duration_box);
  g_signal_connect(G_OBJECT(number_button), "clicked",
		   G_CALLBACK(number_button_clicked), &data2);
  g_signal_connect(G_OBJECT(increment_button), "clicked",
		   G_CALLBACK(increment_button_clicked), &data2);
  g_signal_connect(G_OBJECT(dispense_button), "clicked",
		   G_CALLBACK(dispense_button_clicked), &data4);
  g_signal_connect(G_OBJECT(h_split), "clicked",
		   G_CALLBACK(h_split_clicked), &data4);
  g_signal_connect(G_OBJECT(v_split), "clicked",
		   G_CALLBACK(v_split_clicked), &data4);
  g_signal_connect(G_OBJECT(run_button), "clicked",
		   G_CALLBACK(run_button_clicked), &data);
  g_signal_connect(G_OBJECT(stop_button), "clicked",
		   G_CALLBACK(stop_button_clicked), NULL);
  g_signal_connect(G_OBJECT(resume_button), "clicked",
		   G_CALLBACK(resume_button_clicked), &data);
  g_signal_connect(G_OBJECT(set_voltage_and_frequency_button), "clicked",
		   G_CALLBACK(set_voltage_and_frequency_button_clicked), NULL);

  // signal handlers
  g_signal_connect(G_OBJECT(draw), "expose-event",
		   G_CALLBACK(on_expose_event), main_window);
  g_signal_connect(G_OBJECT(main_window), "button-press-event",
		   G_CALLBACK(mouse_click_event), &data2);
  g_signal_connect(G_OBJECT(main_window), "key-press-event",
		   G_CALLBACK(key_press_event), NULL);
  g_signal_connect(G_OBJECT(main_window), "scroll-event",
		   G_CALLBACK(scroll_callback), NULL);
  g_signal_connect_swapped(G_OBJECT(main_window), "destroy", /* swapped, but WHY? */
			   G_CALLBACK(gtk_main_quit), NULL);
  g_signal_connect(G_OBJECT(repeat_window), "destroy",
		   G_CALLBACK(hide_repeat_window), repeat_window);

  gtk_quit_add(0, clean_up, NULL);
  gtk_widget_show_all(main_window);

  if(configuration.debug()) {
    device.set_debug(true);
  }

  if (feedback_file.is_open()) {
  feedback_file.close();
  }

   if (copy_file.is_open()) {
  copy_file.close();
  }
  // initialization
  initialize(configuration.device_path());

    //initialize the function generator
    CHECK(viOpenDefaultRM(&viDefaultRM)); // &viDefaultRM
    CHECK(viOpen(viDefaultRM, TxtAddress, VI_NULL, VI_NULL, &Instrument)); // 2nd VINULL was 2500

    std::ostringstream reset("*RST\n"); /* Reset the function generator */
	CHECK(viWrite(Instrument, (ViByte*)reset.str().c_str(), (ViUInt32)reset.str().size(), &actual));

    std::ostringstream clear("*CLS\n"); /* Clear errors and status registers */
	CHECK(viWrite(Instrument, (ViByte*)clear.str().c_str(), (ViUInt32)clear.str().size(), &actual));

    std::ostringstream wave("FUNCtion SINusoid\n"); /* Select waveshape */
	CHECK(viWrite(Instrument, (ViByte*)wave.str().c_str(), (ViUInt32)wave.str().size(), &actual));
	/* Other options are SQUare, RAMP, PULSe, NOISe, DC, and USER */

	std::ostringstream output_load("OUTPut:LOAD 50\n"); /* Set the load impedance in Ohms (50 Ohms default) */
	CHECK(viWrite(Instrument, (ViByte*)output_load.str().c_str(), (ViUInt32)output_load.str().size(), &actual));

	//set the values of frequency and voltage from the user
	float initial_volt = atof(gtk_entry_get_text((GtkEntry*)voltageBox));
	read_volt_orig = initial_volt;
	float initial_freq = atof(gtk_entry_get_text((GtkEntry*)frequencyBox));

    std::ostringstream string_init_volt;
    string_init_volt << "VOLTage " << initial_volt << "\n"; /*Set the amplitude in Vpp.  Also see VOLTage:UNIT */
	CHECK(viWrite(Instrument, (ViByte*)string_init_volt.str().c_str(), (ViUInt32)string_init_volt.str().size(), &actual));

    std::ostringstream string_init_freq;
    string_init_freq << "FREQ " << initial_freq << "\n"; /* Set the frequency in Hz) */
	CHECK(viWrite(Instrument, (ViByte*)string_init_freq.str().c_str(), (ViUInt32)string_init_freq.str().size(), &actual));

	//turn on OUTPUT of function generator
	std::ostringstream output_logic("OUTPut ON\n");  /* Turn on the instrument output */
	CHECK(viWrite(Instrument, (ViByte*)output_logic.str().c_str(), (ViUInt32)output_logic.str().size(), &actual));

  //open feedback file for I/O operation
  feedback_file.open ("feedback_file.csv", fstream::in | fstream::out | fstream::app);
  copy_file.open ("copy_file.csv", fstream::in | fstream::out | fstream::app);

  boost::filesystem::path exp_dir = configuration.experiment_logs_dir();
  for(int i=0; true; i++) {
    std::ostringstream experiment_id;
    experiment_id << "experiment " << i;
    gtk_label_set_text(GTK_LABEL(experiment_id_label),
                       experiment_id.str().c_str());
    experiment_id << ".csv";
    // find the first new file (or one that has a zero size
    if(boost::filesystem::exists(exp_dir/experiment_id.str())==false
       || boost::filesystem::file_size(exp_dir/experiment_id.str())==0) {
      device.SetExperimentLogFile((exp_dir/experiment_id.str())
                                  .string().c_str());
      break;
    }
  }

  for(int j=5; j<20; j++) {
    std::ostringstream com_port;
#ifdef _WIN32
    com_port << "COM" << j;
#else
    com_port << "/dev/ttyUSB" << j;
#endif
    string name, version;
    if(device.Connect(com_port.str().c_str())==DmfController::RETURN_OK &&
        device.get_name(name)==DmfController::RETURN_OK &&
        name.compare("Arduino DMF Controller")==0 ) {
      Logging::LogMessage(5, "Connected", NULL, "main");
      device.get_version(version);
      std::vector<float> sampling_rates;
      device.get_available_sampling_rates(sampling_rates);
      device.set_sampling_rate(5);
      string label = name + " " + version;
      gtk_label_set_text(GTK_LABEL(device_info), label.c_str());
      connected = true;
      if(electrodes->size()) {
        update_device();
      }
      break;
    }
  }

  // main loop
  gtk_main();

  return 0;
}
